"""
blink config file.
"""
ALIAS = "blink"

# camera zone names
CABIN_DOORBELL = "cabin doorbell"
DRIVEWAY = "driveway"
BEACH = "beach"
FRONT_YARD = "front yard"
BACK_YARD = "back yard"
WEST = "west"
NORTH = "north"
SOUTH = "south"
NW_SE = "nw-se"

# constants
MAX_HEAT_SETPOINT = 68
MIN_COOL_SETPOINT = 70

MEASUREMENTS = 1  # number of MEASUREMENTS to average
API_TEMPF_MEAN = "temperature_calibrated"

# all environment variables specific to this thermostat type
env_variables = {
    "BLINK_USERNAME": None,
    "BLINK_PASSWORD": None,
    "BLINK_2FA": None,
}

# min required env variables on all runs
required_env_variables = {
    'BLINK_USERNAME': None,
    'BLINK_PASSWORD': None,
    "BLINK_2FA": None,
}

# metadata dict
# 'zone_name' is a placeholder, used at Thermostat class level.
metadata = {
    0: {"zone_name": DRIVEWAY},
    1: {"zone_name": BEACH},
    2: {"zone_name": FRONT_YARD},
    3: {"zone_name": BACK_YARD},
    4: {"zone_name": CABIN_DOORBELL},
    5: {"zone_name": WEST},
    6: {"zone_name": NORTH},
    7: {"zone_name": SOUTH},
    8: {"zone_name": NW_SE},
}

# supported thermostat configs
supported_configs = {"module": "blink",
                     "type": 6,
                     "zones": list(metadata.keys()),
                     "modes": ["OFF_MODE"]}

default_zone = supported_configs["zones"][0]
default_zone_name = ALIAS + "_" + str(default_zone)

argv = [
    "supervise.py",  # module
    ALIAS,  # thermostat
    str(default_zone),  # zone
    "16",  # poll time in sec
    "356",  # reconnect time in sec
    "4",  # tolerance
    "OFF_MODE",  # thermostat mode
    "2",  # number of measurements
    ]



"""KumoCloud integration"""
import os
import time
import traceback

# local imports
from thermostatsupervisor import environment as env
from thermostatsupervisor import kumocloud_config
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import thermostat_common as tc
from thermostatsupervisor import utilities as util

# pykumo import
PYKUMO_DEBUG = False  # debug uses local pykumo repo instead of pkg
if PYKUMO_DEBUG and not env.is_azure_environment():
    mod_path = "..\\pykumo\\pykumo"
    if env.is_interactive_environment():
        mod_path = "..\\" + mod_path
    pykumo = env.dynamic_module_import("pykumo",
                                       mod_path)
else:
    import pykumo  # noqa E402, from path / site packages


class ThermostatClass(pykumo.KumoCloudAccount, tc.ThermostatCommon):
    """KumoCloud thermostat functions."""

    def __init__(self, zone, verbose=True):
        """
        Constructor, connect to thermostat.

        inputs:
            zone(str):  zone of thermostat.
            verbose(bool): debug flag.
        """
        # Kumocloud server auth credentials from env vars
        self.KC_UNAME_KEY = 'KUMO_USERNAME'
        self.KC_PASSWORD_KEY = 'KUMO_PASSWORD'
        self.kc_uname = (os.environ.get(self.KC_UNAME_KEY, "<" +
                         self.KC_UNAME_KEY + "_KEY_MISSING>"))
        self.kc_pwd = (os.environ.get(
            self.KC_PASSWORD_KEY, "<" +
            self.KC_PASSWORD_KEY + "_KEY_MISSING>"))

        # construct the superclass
        # call both parent class __init__
        self._need_fetch = True  # force data fetch
        self.args = [self.kc_uname, self.kc_pwd]
        pykumo.KumoCloudAccount.__init__(self, *self.args)
        tc.ThermostatCommon.__init__(self)

        # set tstat type and debug flag
        self.thermostat_type = kumocloud_config.ALIAS
        self.verbose = verbose

        # configure zone info
        self.zone_number = int(zone)
        self.zone_name = kumocloud_config.metadata[self.zone_number][
            "zone_name"]
        self.device_id = self.get_target_zone_id(self.zone_name)
        self.serial_number = None  # will be populated when unit is queried.
        self.zone_info = None

    def get_target_zone_id(self, zone=0):
        """
        Return the target zone ID.

        inputs:
            zone(int):  zone number.
        returns:
            (int): device_id
        """
        return zone

    def get_zone_index_from_name(self):
        """
        Return zone index for specified zone_name.

        inputs:
            None
        returns:
            (int): zone index.
        """
        if self.verbose:
            print(f"getting index for zone_name={self.zone_name}...")
            print(f"metadata dict={kumocloud_config.metadata}")
        try:
            zone_index = [i for i in kumocloud_config.metadata if
                          kumocloud_config.metadata[i][
                              "zone_name"] == self.zone_name][0]
        except IndexError:
            print(f"ERROR: zone_name={self.zone_name} not present in meta data"
                  f" dict, valid values are {kumocloud_config.metadata.keys()}")
            raise
        return zone_index

    def get_all_metadata(self, zone=None, debug=False):
        """Get all thermostat meta data for zone from kumocloud.

        inputs:
            zone(int): specified zone, if None will print all zones.
            debug(bool): if True will print unit details.
        returns:
            (dict): JSON dict
        """
        return self.get_metadata(zone, None, debug)

    def get_metadata(self, zone=None, parameter=None, debug=False):
        """Get all thermostat meta data for zone from kumocloud.

        inputs:
            zone(int): specified zone, if None will print all zones.
            parameter(str): target parameter, if None will return all.
            debug(bool): if True will print unit details.
        returns:
            (int, float, str, dict): depends on parameter
        """
        try:
            serial_num_lst = list(self.get_indoor_units())  # will query unit
        except UnboundLocalError:  # patch for issue #205
            util.log_msg("WARNING: Kumocloud refresh failed due to "
                         "timeout", mode=util.BOTH_LOG, func_name=1)
            time.sleep(30)
            serial_num_lst = list(self.get_indoor_units())  # retry
        if debug:
            util.log_msg(f"indoor unit serial numbers: {str(serial_num_lst)}",
                         mode=util.DEBUG_LOG + util.CONSOLE_LOG, func_name=1)
        for idx, serial_number in enumerate(serial_num_lst):
            if debug:
                util.log_msg(f"Unit {self.get_name(serial_number)}: address: "
                             f"{self.get_address(serial_number)} credentials: "
                             f"{self.get_credentials(serial_number)}",
                             mode=util.DEBUG_LOG +
                             util.CONSOLE_LOG,
                             func_name=1)
            # populate meta data dict
            if debug:
                print(f"zone index={idx}, serial_number={serial_number}")
            kumocloud_config.metadata[idx]["serial_number"] = serial_number

        # raw_json list:
        # [0]: token, username, device fields
        # [1]: lastupdate date
        # [2]: zone meta data
        # [3]: device token
        if zone is None:
            # returned cached raw data for all zones
            raw_json = self.get_raw_json()[2]  # does not fetch results,
        else:
            # if zone name input, find zone index
            if not isinstance(zone, int):
                zone = self.get_zone_index_from_name()
            # return cached raw data for specified zone, will be a dict
            self.serial_number = serial_num_lst[zone]
            raw_json = self.get_raw_json()[2]['children'][0][
                'zoneTable'][serial_num_lst[zone]]

        if parameter is None:
            return raw_json
        else:
            return raw_json[parameter]

    def print_all_thermostat_metadata(self, zone, debug=False):
        """Print all metadata for zone to the screen.

        inputs:
            zone(int): specified zone, if None will print all zones.
            debug(bool): debug flag
        returns:
            None, prints result to screen
        """
        self.exec_print_all_thermostat_metadata(
            self.get_all_metadata, [zone, debug])


class ThermostatZone(tc.ThermostatCommonZone):
    """
    KumoCloud single zone from kumocloud.

    Class needs to be updated for multi-zone support.
    """

    def __init__(self, Thermostat_obj, verbose=True):
        """
        Zone constructor.

        inputs:
            Thermostat(obj): Thermostat class instance.
            verbose(bool): debug flag.
        """
        # construct the superclass, requires auth setup first
        super().__init__()

        # runtime parameter defaults
        self.poll_time_sec = 10 * 60  # default to 10 minutes
        self.connection_time_sec = 8 * 60 * 60  # default to 8 hours

        # server data cache expiration parameters
        self.fetch_interval_sec = 10  # age of server data before refresh
        self.last_fetch_time = time.time() - 2 * self.fetch_interval_sec

        # switch config for this thermostat
        self.system_switch_position[
            tc.ThermostatCommonZone.HEAT_MODE] = 1  # "Heat" 0 0001
        self.system_switch_position[
            tc.ThermostatCommonZone.OFF_MODE] = 16  # "Off"  1 0000
        self.system_switch_position[
            tc.ThermostatCommonZone.FAN_MODE] = 7  # "Fan"   0 0111
        self.system_switch_position[
            tc.ThermostatCommonZone.DRY_MODE] = 2  # dry     0 0010
        self.system_switch_position[
            tc.ThermostatCommonZone.COOL_MODE] = 3  # cool   0 0011
        self.system_switch_position[
            tc.ThermostatCommonZone.AUTO_MODE] = 33  # auto   0 0101

        # zone info
        self.verbose = verbose
        self.thermostat_type = kumocloud_config.ALIAS
        self.device_id = Thermostat_obj.device_id
        self.Thermostat = Thermostat_obj
        self.zone_info = Thermostat_obj.get_all_metadata(
            Thermostat_obj.zone_number)
        self.zone_number = Thermostat_obj.zone_number
        self.zone_name = self.get_zone_name()  # get real zone name from device

    def get_parameter(self, key, parent_key=None, grandparent_key=None,
                      default_val=None):
        """
        Get parameter from zone dictionary.

        inputs:
            key(str): target dict key
            parent_key(str): first level dict key
            grandparent_key(str): second level dict key
            default_val(str, int, float): default value on key errors
        """
        return_val = default_val
        if grandparent_key is not None:
            try:
                # check parent keys
                grandparent_dict = self.zone_info[grandparent_key]
                parent_dict = grandparent_dict[parent_key]
                return_val = parent_dict[key]
            except KeyError:
                if default_val is None:
                    # if no default val, then display detailed key error
                    util.log_msg(traceback.format_exc(),
                                 mode=util.BOTH_LOG, func_name=1)
                    util.log_msg(f"raw zone_info dict={self.zone_info}",
                                 mode=util.BOTH_LOG, func_name=1)
                    raise
        elif parent_key is not None:
            try:
                parent_dict = self.zone_info[parent_key]
                return_val = parent_dict[key]
            except KeyError:
                if default_val is None:
                    # if no default val, then display detailed key error
                    util.log_msg(traceback.format_exc(),
                                 mode=util.BOTH_LOG, func_name=1)
                    util.log_msg(f"raw zone_info dict={self.zone_info}",
                                 mode=util.BOTH_LOG, func_name=1)
                    raise
        else:
            try:
                return_val = self.zone_info[key]
            except (KeyError, TypeError):
                if default_val is None:
                    # if no default val, then display detailed key error
                    util.log_msg(traceback.format_exc(),
                                 mode=util.BOTH_LOG, func_name=1)
                    util.log_msg(f"target key={key}, "
                                 f"raw zone_info dict={self.zone_info}",
                                 mode=util.BOTH_LOG, func_name=1)
                    raise
        return return_val

    def get_zone_name(self):
        """
        Return the name associated with the zone number.

        inputs:
            None
        returns:
            (str) zone name
        """
        self.refresh_zone_info()
        zone_name = self.get_parameter('label')
        # update metadata dict.
        kumocloud_config.metadata[self.zone_number]["zone_name"] = zone_name
        return zone_name

    def get_display_temp(self) -> float:  # used
        """
        Refresh the cached zone information and return Indoor Temp in Deg F.

        inputs:
            None
        returns:
            (float): indoor temp in deg F.
        """
        self.refresh_zone_info()
        return util.c_to_f(self.get_parameter('room_temp',
                                              'reportedCondition',
                                              default_val=util.BOGUS_INT))

    def get_display_humidity(self) -> (float, None):
        """
        Refresh the cached zone information and return IndoorHumidity.

        inputs:
            None
        returns:
            (float, None): indoor humidity in %RH, None if not supported.
        """
        if not self.get_is_humidity_supported():
            return None
        else:
            # untested, don't have humidity support
            # zone refreshed during if clause above
            return util.c_to_f(self.get_parameter('humidity',
                                                  'reportedCondition',
                                                  default_val=util.BOGUS_INT))

    def get_is_humidity_supported(self) -> bool:  # used
        """
        Refresh the cached zone information and return the
          True if humidity sensor data is trustworthy.

        inputs:
            None
        returns:
            (booL): True if is in humidity sensor is available and not faulted.
        """
        self.refresh_zone_info()
        return self.get_parameter('humidistat', 'inputs',
                                  'acoilSettings')

    def is_heat_mode(self) -> int:
        """
        Refresh the cached zone information and return the heat mode.

        inputs:
            None
        returns:
            (int) heat mode, 1=enabled, 0=disabled.
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.HEAT_MODE])

    def is_cool_mode(self) -> int:
        """
        Refresh the cached zone information and return the cool mode.

        inputs:
            None
        returns:
            (int): cool mode, 1=enabled, 0=disabled.
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.COOL_MODE])

    def is_dry_mode(self) -> int:
        """
        Refresh the cached zone information and return the dry mode.

        inputs:
            None
        returns:
            (int): dry mode, 1=enabled, 0=disabled.
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.DRY_MODE])

    def is_fan_mode(self) -> int:
        """
        Refresh the cached zone information and return the fan mode.

        inputs:
            None
        returns:
            (int): fan mode, 1=enabled, 0=disabled.
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.FAN_MODE])

    def is_auto_mode(self) -> int:
        """
        Refresh the cached zone information and return the auto mode.

        inputs:
            None
        returns:
            (int): auto mode, 1=enabled, 0=disabled.
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.AUTO_MODE])

    def is_off_mode(self) -> int:
        """
        Refresh the cached zone information and return the off mode.

        inputs:
            None
        returns:
            (int): off mode, 1=enabled, 0=disabled.
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.OFF_MODE])

    def is_heating(self):
        """Return 1 if heating relay is active, else 0."""
        return int(self.is_heat_mode() and self.is_power_on() and
                   self.get_heat_setpoint_raw() > self.get_display_temp())

    def is_cooling(self):
        """Return 1 if cooling relay is active, else 0."""
        return int(self.is_cool_mode() and self.is_power_on() and
                   self.get_cool_setpoint_raw() < self.get_display_temp())

    def is_drying(self):
        """Return 1 if drying relay is active, else 0."""
        return int(self.is_dry_mode() and self.is_power_on() and
                   self.get_cool_setpoint_raw() < self.get_display_temp())

    def is_auto(self):
        """Return 1 if auto relay is active, else 0."""
        return int(self.is_auto_mode() and self.is_power_on() and
                   (self.get_cool_setpoint_raw() < self.get_display_temp() or
                    self.get_heat_setpoint_raw() > self.get_display_temp()))

    def is_fanning(self):
        """Return 1 if fan relay is active, else 0."""
        return int(self.is_fan_on() and self.is_power_on())

    def is_power_on(self):
        """Return 1 if power relay is active, else 0."""
        self.refresh_zone_info()
        return self.get_parameter('power', 'reportedCondition', default_val=0)

    def is_fan_on(self):
        """Return 1 if fan relay is active, else 0."""
        if self.is_power_on():
            fan_speed = self.get_parameter('fan_speed', 'reportedCondition')
            if fan_speed is None:
                return 0  # no fan_speed key, return 0
            else:
                return int(fan_speed > 0 or
                           self.get_parameter('fan_speed_text', 'more',
                                              'reportedCondition') != 'off')
        else:
            return 0

    def is_defrosting(self):
        """Return 1 if defrosting is active, else 0."""
        self.refresh_zone_info()
        return int(self.get_parameter('defrost', 'status_display',
                                      'reportedCondition'))

    def is_standby(self):
        """Return 1 if standby is active, else 0."""
        self.refresh_zone_info()
        return int(self.get_parameter('standby', 'status_display',
                                      'reportedCondition'))

    def get_heat_setpoint_raw(self) -> int:  # used
        """
        Refresh the cached zone information and return the heat setpoint.

        inputs:
            None
        returns:
            (int): heating set point in degrees F.
        """
        self.refresh_zone_info()
        # if power is off then sp_heat may be missing
        return util.c_to_f(self.get_parameter('sp_heat',
                                              'reportedCondition',
                                              default_val=-1))

    def get_heat_setpoint(self) -> str:
        """Return heat setpoint with units as a string."""
        return util.temp_value_with_units(self.get_heat_setpoint_raw())

    def get_schedule_heat_sp(self) -> int:  # used
        """
        Return the schedule heat setpoint.

        inputs:
            None
        returns:
            (int): scheduled heating set point in degrees.
        """
        return kumocloud_config.MAX_HEAT_SETPOINT  # max heat set point allowed

    def get_schedule_cool_sp(self) -> int:
        """
        Return the schedule cool setpoint.

        inputs:
            None
        returns:
            (int): scheduled cooling set point in degrees F.
        """
        return kumocloud_config.MIN_COOL_SETPOINT  # min cool set point allowed

    def get_cool_setpoint_raw(self) -> int:
        """
        Return the cool setpoint.

        inputs:
            None
        returns:
            (int): cooling set point in degrees F.
        """
        self.refresh_zone_info()
        # if power is off then sp_heat may be missing
        return util.c_to_f(self.get_parameter('sp_cool',
                                              'reportedCondition',
                                              default_val=-1))

    def get_cool_setpoint(self) -> str:
        """Return cool setpoint with units as a string."""
        return util.temp_value_with_units(self.get_cool_setpoint_raw())

    def get_is_invacation_hold_mode(self) -> bool:  # used
        """
        Return the
          'IsInVacationHoldMode' setting.

        inputs:
            None
        returns:
            (booL): True if is in vacation hold mode.
        """
        return False  # no schedule, hold not implemented

    def get_vacation_hold(self) -> bool:
        """
        Return the
        VacationHold setting.

        inputs:
            None
        returns:
            (bool): True if vacation hold is set.
        """
        return False  # no schedule, hold not implemented

    def get_system_switch_position(self) -> int:  # used
        """
        Return the system switch position.

        inputs:
            None
        returns:
            (int) current mode for unit, should match value
                  in self.system_switch_position
        """
        self.refresh_zone_info()
        # first check if power is on
        # if power is off then operation_mode key may be missing.
        if not self.is_power_on():
            return self.system_switch_position[
                tc.ThermostatCommonZone.OFF_MODE]
        else:
            return self.get_parameter('operation_mode',
                                      'reportedCondition')

    def set_heat_setpoint(self, temp: int) -> None:
        """
        Set a new heat setpoint.

        This will also attempt to turn the thermostat to 'Heat'
        inputs:
            temp(int): desired temperature in F
        returns:
            None
        """
        # self.device_id.set_heat_setpoint(self._f_to_c(temp))
        # TODO needs implementation
        del temp
        util.log_msg("WARNING: this method not implemented yet for this "
                     "thermostat type",
                     mode=util.BOTH_LOG, func_name=1)

    def set_cool_setpoint(self, temp: int) -> None:
        """
        Set a new cool setpoint.

        This will also attempt to turn the thermostat to 'Cool'
        inputs:
            temp(int): desired temperature in deg F.
        returns:
            None
        """
        # self.device_id.set_cool_setpoint(self._f_to_c(temp))
        # TODO needs implementation
        del temp
        util.log_msg("WARNING: this method not implemented yet for this "
                     "thermostat type",
                     mode=util.BOTH_LOG, func_name=1)

    def refresh_zone_info(self, force_refresh=False):
        """
        Refresh zone info from KumoCloud.

        inputs:
            force_refresh(bool): if True, ignore expiration timer.
        returns:
            None, zone_data is refreshed.
        """
        now_time = time.time()
        # refresh if past expiration date or force_refresh option
        if (force_refresh or (now_time >=
                              (self.last_fetch_time +
                               self.fetch_interval_sec))):
            self.Thermostat._need_fetch = True \
                # pylint: disable=protected-access
            try:
                self.Thermostat._fetch_if_needed() \
                    # pylint: disable=protected-access
            except UnboundLocalError:  # patch for issue #205
                util.log_msg("WARNING: Kumocloud refresh failed due to "
                             "timeout", mode=util.BOTH_LOG, func_name=1)
            self.last_fetch_time = now_time
            # refresh device object
            self.zone_info = self.Thermostat.get_all_metadata(
                self.zone_number)

    def report_heating_parameters(self, switch_position=None):
        """
        Display critical thermostat settings and reading to the screen.

        inputs:
            switch_position(int): switch position override, used for testing.
        returns:
            None
        """
        # current temp as measured by thermostat
        util.log_msg(f"display temp="
                     f"{util.temp_value_with_units(self.get_display_temp())}",
                     mode=util.BOTH_LOG,
                     func_name=1)

        # get switch position
        if switch_position is None:
            switch_position = self.get_system_switch_position()

        # heating status
        if switch_position == \
                self.system_switch_position[self.HEAT_MODE]:
            util.log_msg(f"heat mode={self.is_heat_mode()}",
                         mode=util.BOTH_LOG)
            util.log_msg(
                f"heat setpoint={self.get_heat_setpoint_raw()}",
                mode=util.BOTH_LOG)
            util.log_msg(
                f"schedule heat sp={self.get_schedule_heat_sp()}",
                mode=util.BOTH_LOG)

        # cooling status
        if switch_position == \
                self.system_switch_position[self.COOL_MODE]:
            util.log_msg(f"cool mode={self.is_cool_mode()}",
                         mode=util.BOTH_LOG)
            util.log_msg(
                f"cool setpoint={self.get_cool_setpoint_raw()}",
                mode=util.BOTH_LOG)
            util.log_msg(
                f"schedule cool sp={self.get_schedule_cool_sp()}",
                mode=util.BOTH_LOG)

        # hold settings
        util.log_msg(
            f"is in vacation hold mode={self.get_is_invacation_hold_mode()}",
            mode=util.BOTH_LOG)
        util.log_msg(f"vacation hold={self.get_vacation_hold()}",
                     mode=util.BOTH_LOG)
        util.log_msg(
            f"vacation hold until time={self.get_vacation_hold_until_time()}",
            mode=util.BOTH_LOG)
        util.log_msg(f"temporary hold until time="
                     f"{self.get_temporary_hold_until_time()}",
                     mode=util.BOTH_LOG)


if __name__ == "__main__":

    # verify environment
    env.get_python_version()

    # get zone override
    api.uip = api.UserInputs(argv_list=None,
                             thermostat_type=kumocloud_config.ALIAS)
    zone_number = api.uip.get_user_inputs(api.uip.zone_name,
                                          api.input_flds.zone)

    tc.thermostat_basic_checkout(
        kumocloud_config.ALIAS,
        zone_number,
        ThermostatClass, ThermostatZone)

    tc.thermostat_get_all_zone_temps(
        kumocloud_config.ALIAS,
        kumocloud_config.supported_configs["zones"],
        ThermostatClass,
        ThermostatZone)



"""
Honeywell thermostat config file.
"""
ALIAS = "honeywell"

# constants

# all environment variables specific to this thermostat type
env_variables = {
    "TCC_USERNAME": None,
    "TCC_PASSWORD": None,
}

# min required env variables on all runs
required_env_variables = {
    "TCC_USERNAME": None,
    "TCC_PASSWORD": None,
}

# supported thermostat configs
supported_configs = {"module": "honeywell",
                     "type": 1,
                     "zones": [0],
                     "modes": ["OFF_MODE", "HEAT_MODE", "COOL_MODE",
                               "UNKNOWN_MODE"]}
default_zone = supported_configs["zones"][0]
default_zone_name = ALIAS + "_" + str(default_zone)

argv = [
    "supervise.py",  # module
    ALIAS,  # thermostat
    str(default_zone),  # zone
    "19",  # poll time in sec
    "359",  # reconnect time in sec
    "3",  # tolerance
    "OFF_MODE",  # thermostat mode
    "2",  # number of measurements
    ]



"""KumoCloud integration using local API for data."""
import os
import time

# local imports
from thermostatsupervisor import kumolocal_config
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import thermostat_common as tc
from thermostatsupervisor import environment as env
from thermostatsupervisor import utilities as util

# pykumo
PYKUMO_DEBUG = False  # debug uses local pykumo repo instead of pkg
if PYKUMO_DEBUG and not env.is_azure_environment():
    mod_path = "..\\pykumo\\pykumo"
    if env.is_interactive_environment():
        mod_path = "..\\" + mod_path
    pykumo = env.dynamic_module_import("pykumo",
                                       mod_path)
else:
    import pykumo  # noqa E402, from path / site packages


class ThermostatClass(pykumo.KumoCloudAccount, tc.ThermostatCommon):
    """KumoCloud thermostat functions."""

    def __init__(self, zone, verbose=True):
        """
        Constructor, connect to thermostat.

        inputs:
            zone(str):  zone of thermostat.
            verbose(bool): debug flag.
        """
        # Kumocloud server auth credentials from env vars
        self.KC_UNAME_KEY = 'KUMO_USERNAME'
        self.KC_PASSWORD_KEY = 'KUMO_PASSWORD'
        self.kc_uname = (os.environ.get(self.KC_UNAME_KEY, "<" +
                         self.KC_UNAME_KEY + "_KEY_MISSING>"))
        self.kc_pwd = (os.environ.get(
            self.KC_PASSWORD_KEY, "<" +
            self.KC_PASSWORD_KEY + "_KEY_MISSING>"))

        # construct the superclass
        # call both parent class __init__
        self.args = [self.kc_uname, self.kc_pwd]
        pykumo.KumoCloudAccount.__init__(self, *self.args)
        tc.ThermostatCommon.__init__(self)

        # set tstat type and debug flag
        self.thermostat_type = kumolocal_config.ALIAS
        self.verbose = verbose

        # configure zone info
        self.zone_number = int(zone)
        self.zone_name = self.get_zone_name()
        self.device_id = None  # initialize
        self.device_id = self.get_target_zone_id(self.zone_number)
        self.serial_number = None  # will be populated when unit is queried.

    def get_zone_name(self):
        """
        Return the name associated with the zone number from metadata dict.

        inputs:
            None
        returns:
            (str) zone name
        """
        return kumolocal_config.metadata[self.zone_number]["zone_name"]

    def get_target_zone_id(self, zone=0):
        """
        Return the target zone ID.

        inputs:
            zone(int): zone number.
        returns:
            (obj): PyKumo object
        """
        # populate the zone dictionary
        # establish local interface to kumos, must be on local net
        kumos = self.make_pykumos()
        device_id = kumos[self.zone_name]
        # print zone name the first time it is known
        if self.device_id is None:
            util.log_msg(
                f"zone {zone} name = '{self.zone_name}', "
                f"device_id={device_id}",
                mode=util.DEBUG_LOG +
                util.CONSOLE_LOG,
                func_name=1)
        self.device_id = device_id

        # return the target zone object
        return self.device_id

    def get_kumocloud_thermostat_metadata(self, zone=None, debug=False):
        """Get all thermostat meta data for zone from kumocloud.

        inputs:
            zone(): specified zone, if None will print all zones.
            debug(bool): debug flag.
        returns:
            (dict): JSON dict
        """
        del debug  # unused
        try:
            units = list(self.get_indoor_units())  # will also query unit
        except UnboundLocalError:  # patch for issue #205
            util.log_msg("WARNING: Kumocloud refresh failed due to "
                         "timeout", mode=util.BOTH_LOG, func_name=1)
            time.sleep(10)
            units = list(self.get_indoor_units())  # retry
        util.log_msg(f"indoor unit serial numbers: {str(units)}",
                     mode=util.DEBUG_LOG + util.CONSOLE_LOG, func_name=1)
        for serial_number in units:
            util.log_msg(
                f"Unit {self.get_name(serial_number)}: address: "
                f"{self.get_address(serial_number)} credentials: "
                f"{self.get_credentials(serial_number)}",
                mode=util.DEBUG_LOG +
                util.CONSOLE_LOG,
                func_name=1)
        if zone is None:
            # returned cached raw data for all zones
            raw_json = self.get_raw_json()  # does not fetch results,
        else:
            # return cached raw data for specified zone
            self.serial_number = units[zone]
            raw_json = self.get_raw_json()[2]['children'][0][
                'zoneTable'][units[zone]]
        return raw_json

    def get_all_metadata(self, zone=None, debug=False):
        """Get all thermostat meta data for device_id from local API.

        inputs:
            zone(): specified zone
            debug(bool): debug flag.
        returns:
            (dict): dictionary of meta data.
        """
        return self.get_metadata(zone, None, debug)

    def get_metadata(self, zone=None, parameter=None, debug=False):
        """Get thermostat meta data for device_id from local API.

        inputs:
            zone(str or int): (unused) specified zone
            parameter(str): target parameter, if None will return all.
            debug(bool): debug flag.
        returns:
            (dict): dictionary of meta data.
        """
        del debug  # unused
        del zone  # unused

        # refresh device status
        self.device_id.update_status()
        meta_data = {}
        meta_data['status'] = self.device_id.get_status()
        # pylint: disable=protected-access
        meta_data['sensors'] = self.device_id._sensors
        # pylint: disable=protected-access
        meta_data['profile'] = self.device_id._profile
        if parameter is None:
            return meta_data
        else:
            return meta_data[parameter]

    def print_all_thermostat_metadata(self, zone, debug=False):
        """Print all metadata for zone to the screen.

        inputs:
            zone(int): specified zone, if None will print all zones.
            debug(bool): debug flag
        returns:
            None, prints result to screen
        """
        self.exec_print_all_thermostat_metadata(
            self.get_all_metadata, [zone, debug])


class ThermostatZone(tc.ThermostatCommonZone):
    """
    KumoCloud single zone on local network.

    Class needs to be updated for multi-zone support.
    """

    def __init__(self, Thermostat_obj, verbose=True):
        """
        Zone constructor.

        inputs:
            Thermostat(obj): Thermostat class instance.
            verbose(bool): debug flag.
        """
        # construct the superclass, requires auth setup first
        super().__init__()

        # runtime parameter defaults
        self.poll_time_sec = 10 * 60  # default to 10 minutes
        self.connection_time_sec = 8 * 60 * 60  # default to 8 hours

        # server data cache expiration parameters
        self.fetch_interval_sec = 10  # age of server data before refresh
        self.last_fetch_time = time.time() - 2 * self.fetch_interval_sec

        # switch config for this thermostat
        self.system_switch_position[tc.ThermostatCommonZone.COOL_MODE] = "cool"
        self.system_switch_position[tc.ThermostatCommonZone.HEAT_MODE] = "heat"
        self.system_switch_position[tc.ThermostatCommonZone.OFF_MODE] = "off"
        self.system_switch_position[tc.ThermostatCommonZone.DRY_MODE] = "dry"
        self.system_switch_position[tc.ThermostatCommonZone.AUTO_MODE] = "auto"
        self.system_switch_position[tc.ThermostatCommonZone.FAN_MODE] = "vent"

        # zone info
        self.verbose = verbose
        self.thermostat_type = kumolocal_config.ALIAS
        self.device_id = Thermostat_obj.device_id
        self.Thermostat = Thermostat_obj
        self.zone_number = Thermostat_obj.zone_number
        self.zone_name = Thermostat_obj.zone_name
        self.zone_name = self.get_zone_name()

    def get_zone_name(self):
        """
        Return the name associated with the zone number from device memory.

        inputs:
            None
        returns:
            (str) zone name
        """
        self.refresh_zone_info()
        zone_name = self.device_id.get_name()
        # update metadata dict
        kumolocal_config.metadata[self.zone_number]["zone_name"] = zone_name
        return zone_name

    def get_display_temp(self) -> float:  # used
        """
        Refresh the cached zone information and return Indoor Temp in Deg F.

        inputs:
            None
        returns:
            (float): indoor temp in deg F.
        """
        self.refresh_zone_info()
        return util.c_to_f(self.device_id.get_current_temperature())

    def get_display_humidity(self) -> (float, None):
        """
        Refresh the cached zone information and return IndoorHumidity.

        inputs:
            None
        returns:
            (float, None): indoor humidity in %RH, None if not supported.
        """
        self.refresh_zone_info()
        return self.device_id.get_current_humidity()

    def get_is_humidity_supported(self) -> bool:  # used
        """
        Refresh the cached zone information and return the
          True if humidity sensor data is trustworthy.

        inputs:
            None
        returns:
            (booL): True if is in humidity sensor is available and not faulted.
        """
        return self.get_display_humidity() is not None

    def is_heat_mode(self) -> int:
        """
        Refresh the cached zone information and return the heat mode.

        inputs:
            None
        returns:
            (int) heat mode, 1=enabled, 0=disabled.
        """
        self.refresh_zone_info()
        return int(self.device_id.get_mode() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.HEAT_MODE])

    def is_cool_mode(self) -> int:
        """
        Refresh the cached zone information and return the cool mode.

        inputs:
            None
        returns:
            (int): cool mode, 1=enabled, 0=disabled.
        """
        self.refresh_zone_info()
        return int(self.device_id.get_mode() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.COOL_MODE])

    def is_dry_mode(self) -> int:
        """
        Refresh the cached zone information and return the dry mode.

        inputs:
            None
        returns:
            (int): dry mode, 1=enabled, 0=disabled.
        """
        self.refresh_zone_info()
        return int(self.device_id.get_mode() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.DRY_MODE])

    def is_fan_mode(self) -> int:
        """
        Refresh the cached zone information and return the fan mode.

        inputs:
            None
        returns:
            (int): fan mode, 1=enabled, 0=disabled.
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.FAN_MODE])

    def is_auto_mode(self) -> int:
        """
        Refresh the cached zone information and return the auto mode.

        inputs:
            None
        returns:
            (int): auto mode, 1=enabled, 0=disabled.
        """
        self.refresh_zone_info()
        return int(self.device_id.get_mode() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.AUTO_MODE])

    def is_off_mode(self) -> int:
        """
        Refresh the cached zone information and return the off mode.

        inputs:
            None
        returns:
            (int): off mode, 1=enabled, 0=disabled.
        """
        return int(self.get_system_switch_position() ==
                   self.system_switch_position[
                       tc.ThermostatCommonZone.OFF_MODE])

    def is_heating(self):
        """Return 1 if heating relay is active, else 0."""
        return int(self.is_heat_mode() and self.is_power_on() and
                   self.get_heat_setpoint_raw() > self.get_display_temp())

    def is_cooling(self):
        """Return 1 if cooling relay is active, else 0."""
        return int(self.is_cool_mode() and self.is_power_on() and
                   self.get_cool_setpoint_raw() < self.get_display_temp())

    def is_drying(self):
        """Return 1 if drying relay is active, else 0."""
        return int(self.is_dry_mode() and self.is_power_on() and
                   self.get_cool_setpoint_raw() < self.get_display_temp())

    def is_auto(self):
        """Return 1 if auto relay is active, else 0."""
        return int(self.is_auto_mode() and self.is_power_on() and
                   (self.get_cool_setpoint_raw() < self.get_display_temp() or
                    self.get_heat_setpoint_raw() > self.get_display_temp()))

    def is_fanning(self):
        """Return 1 if fan relay is active, else 0."""
        return int(self.is_fan_on() and self.is_power_on())

    def is_power_on(self):
        """Return 1 if power relay is active, else 0."""
        self.refresh_zone_info()
        return int(self.device_id.get_mode() != 'off')

    def is_fan_on(self):
        """Return 1 if fan relay is active, else 0."""
        self.refresh_zone_info()
        return int(self.device_id.get_fan_speed() != 'off')

    def is_defrosting(self):
        """Return 1 if defrosting is active, else 0."""
        self.refresh_zone_info()
        return int(self.device_id.get_status('defrost') == "True")

    def is_standby(self):
        """Return 1 if standby is active, else 0."""
        self.refresh_zone_info()
        return int(self.device_id.get_standby())

    def get_heat_setpoint_raw(self) -> int:  # used
        """
        Refresh the cached zone information and return the heat setpoint.

        inputs:
            None
        returns:
            (int): heating set point in degrees F.
        """
        self.refresh_zone_info()
        return util.c_to_f(self.device_id.get_heat_setpoint())

    def get_heat_setpoint(self) -> str:
        """Return heat setpoint with units as a string."""
        return util.temp_value_with_units(self.get_heat_setpoint_raw())

    def get_schedule_heat_sp(self) -> int:  # used
        """
        Return the schedule heat setpoint.

        inputs:
            None
        returns:
            (int): scheduled heating set point in degrees.
        """
        return kumolocal_config.MAX_HEAT_SETPOINT  # max heat set point allowed

    def get_schedule_cool_sp(self) -> int:
        """
        Return the schedule cool setpoint.

        inputs:
            None
        returns:
            (int): scheduled cooling set point in degrees F.
        """
        return kumolocal_config.MIN_COOL_SETPOINT  # min cool set point allowed

    def get_cool_setpoint_raw(self) -> int:
        """
        Return the cool setpoint.

        inputs:
            None
        returns:
            (int): cooling set point in degrees F.
        """
        self.refresh_zone_info()
        return util.c_to_f(self.device_id.get_cool_setpoint())

    def get_cool_setpoint(self) -> str:
        """Return cool setpoint with units as a string."""
        return util.temp_value_with_units(self.get_cool_setpoint_raw())

    def get_is_invacation_hold_mode(self) -> bool:  # used
        """
        Return the
          'IsInVacationHoldMode' setting.

        inputs:
            None
        returns:
            (booL): True if is in vacation hold mode.
        """
        return False  # no schedule, hold not implemented

    def get_vacation_hold(self) -> bool:
        """
        Return the
        VacationHold setting.

        inputs:
            None
        returns:
            (bool): True if vacation hold is set.
        """
        return False  # no schedule, hold not implemented

    def get_system_switch_position(self) -> int:  # used
        """
        Return the system switch position, same as mode.

        inputs:
            None
        returns:
            (int) current mode for unit, should match value
                  in self.system_switch_position
        """
        self.refresh_zone_info()
        return self.device_id.get_mode()

    def set_heat_setpoint(self, temp: int) -> None:
        """
        Set a new heat setpoint.

        This will also attempt to turn the thermostat to 'Heat'
        inputs:
            temp(int): desired temperature in F
        returns:
            None
        """
        self.device_id.set_heat_setpoint(util.f_to_c(temp))

    def set_cool_setpoint(self, temp: int) -> None:
        """
        Set a new cool setpoint.

        This will also attempt to turn the thermostat to 'Cool'
        inputs:
            temp(int): desired temperature in deg F.
        returns:
            None
        """
        self.device_id.set_cool_setpoint(util.f_to_c(temp))

    def refresh_zone_info(self, force_refresh=False):
        """
        Refresh zone info from KumoCloud.

        inputs:
            force_refresh(bool): if True, ignore expiration timer.
        returns:
            None, device_id object is refreshed.
        """
        now_time = time.time()
        # refresh if past expiration date or force_refresh option
        if (force_refresh or (now_time >=
                              (self.last_fetch_time +
                               self.fetch_interval_sec))):
            self.Thermostat._need_fetch = True \
                # pylint: disable=protected-access
            try:
                self.Thermostat._fetch_if_needed() \
                    # pylint: disable=protected-access
            except UnboundLocalError:  # patch for issue #205
                util.log_msg("WARNING: Kumocloud refresh failed due to "
                             "timeout", mode=util.BOTH_LOG, func_name=1)
            self.last_fetch_time = now_time
            # refresh device object
            self.device_id = \
                self.Thermostat.get_target_zone_id(self.zone_name)

    def report_heating_parameters(self, switch_position=None):
        """
        Display critical thermostat settings and reading to the screen.

        inputs:
            switch_position(int): switch position override, used for testing.
        returns:
            None
        """
        # current temp as measured by thermostat
        util.log_msg(
            f"display temp="
            f"{util.temp_value_with_units(self.get_display_temp())}",
            mode=util.BOTH_LOG,
            func_name=1)

        # get switch position
        if switch_position is None:
            switch_position = self.get_system_switch_position()

        # heating status
        if switch_position == \
                self.system_switch_position[self.HEAT_MODE]:
            util.log_msg(f"heat mode={self.is_heat_mode()}",
                         mode=util.BOTH_LOG)
            util.log_msg(
                f"heat setpoint={self.get_heat_setpoint_raw()}",
                mode=util.BOTH_LOG)
            util.log_msg(
                f"schedule heat sp={self.get_schedule_heat_sp()}",
                mode=util.BOTH_LOG)

        # cooling status
        if switch_position == \
                self.system_switch_position[self.COOL_MODE]:
            util.log_msg(f"cool mode={self.is_cool_mode()}",
                         mode=util.BOTH_LOG)
            util.log_msg(
                f"cool setpoint={self.get_cool_setpoint_raw()}",
                mode=util.BOTH_LOG)
            util.log_msg(
                f"schedule cool sp={self.get_schedule_cool_sp()}",
                mode=util.BOTH_LOG)

        # hold settings
        util.log_msg(
            f"is in vacation hold mode={self.get_is_invacation_hold_mode()}",
            mode=util.BOTH_LOG)
        util.log_msg(f"vacation hold={self.get_vacation_hold()}",
                     mode=util.BOTH_LOG)
        util.log_msg(
            f"vacation hold until time={self.get_vacation_hold_until_time()}",
            mode=util.BOTH_LOG)
        util.log_msg(
            f"temporary hold until time="
            f"{self.get_temporary_hold_until_time()}",
            mode=util.BOTH_LOG)


if __name__ == "__main__":

    # verify environment
    env.get_python_version()

    # get zone override
    api.uip = api.UserInputs(argv_list=None,
                             thermostat_type=kumolocal_config.ALIAS)
    zone_number = api.uip.get_user_inputs(api.uip.zone_name,
                                          api.input_flds.zone)

    tc.thermostat_basic_checkout(
        kumolocal_config.ALIAS,
        zone_number,
        ThermostatClass, ThermostatZone)

    tc.thermostat_get_all_zone_temps(
        kumolocal_config.ALIAS,
        kumolocal_config.supported_configs["zones"],
        ThermostatClass,
        ThermostatZone)



"""Environment variable handling."""

# built-in libraries
import importlib.util
import os
import socket
import sys
import traceback

# third party libraries
import psutil

# local imports
from thermostatsupervisor import utilities as util

# thermostat config files
from thermostatsupervisor import honeywell_config
from thermostatsupervisor import kumocloud_config
from thermostatsupervisor import kumolocal_config
from thermostatsupervisor import mmm_config
from thermostatsupervisor import sht31_config

MIN_PYTHON_MAJOR_VERSION = 3  # minimum python major version required
MIN_PYTHON_MINOR_VERSION = 7  # minimum python minor version required

# all environment variables required by code should be registered here
env_variables = {
    "GMAIL_USERNAME": None,
    "GMAIL_PASSWORD": None,
}
env_variables.update(honeywell_config.env_variables)
env_variables.update(kumocloud_config.env_variables)
env_variables.update(kumolocal_config.env_variables)
env_variables.update(mmm_config.env_variables)
env_variables.update(sht31_config.env_variables)


def get_env_variable(env_key):
    """
    Get environment variable.

    Results will be logged but passwords will be masked off.

    inputs:
       env_key(str): env variable of interest
       debug(bool): verbose debugging
    returns:
       (dict): {status, value, key}
    """
    # defaults
    return_buffer = {
        "status": util.NO_ERROR,
        "value": None,
        "key": env_key,
    }

    try:
        # unit test key is not required to be in env var list
        if env_key == sht31_config.UNIT_TEST_ENV_KEY:
            return_buffer["value"] = get_local_ip()
        else:
            return_buffer["value"] = os.environ[env_key]

        # mask off any password keys
        if "PASSWORD" in return_buffer["key"]:
            value_shown = "(hidden)"
        else:
            value_shown = return_buffer["value"]

        util.log_msg(f"{env_key}={value_shown}",
                     mode=util.DEBUG_LOG)
    except KeyError:
        util.log_msg(f"FATAL ERROR: required environment variable '{env_key}'"
                     " is missing.", mode=util.CONSOLE_LOG + util.DATA_LOG)
        return_buffer["status"] = util.ENVIRONMENT_ERROR
    return return_buffer


def load_all_env_variables():
    """
    Load all environment variables into a dictionary.

    inputs:
        None
    returns:
        None, populates env_variables dict.
    """
    for key in env_variables:
        util.log_msg(f"checking key: {key}",
                     mode=util.BOTH_LOG, func_name=1)
        env_variables[key] = get_env_variable(key)["value"]


def is_interactive_environment():
    """Return True if script is run through IDE."""
    parent = psutil.Process(os.getpid()).parent().name()
    if parent in ["cmd.exe", "py.exe", "bash"]:
        return False
    elif parent == "eclipse.exe":
        return True
    else:
        print(f"DEBUG: parent process={parent}")
        raise OSError(f"unrecognized environment: {parent}")


def get_local_ip():
    """Get local IP address for this PC."""
    socket_obj = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        socket_obj.connect(('10.255.255.255', 1))
        ip_address = socket_obj.getsockname()[0]
    except Exception:
        util.log_msg(traceback.format_exc(),
                     mode=util.BOTH_LOG, func_name=1)
        ip_address = '127.0.0.1'
    finally:
        socket_obj.close()
    return ip_address


def is_azure_environment():
    """
    Return True if machine is Azure pipeline.

    Function assumes '192.' IP addresses are not Azure,
    everything else is Azure.
    """
    return '192.' not in get_local_ip()


def get_python_version(min_major_version=MIN_PYTHON_MAJOR_VERSION,
                       min_minor_version=MIN_PYTHON_MINOR_VERSION,
                       display_version=True):
    """
    Print current Python version to the screen.

    inputs:
        min_major_version(int): min allowed major version
        min_minor_version(int): min allowed minor version
        display_version(bool): True to print to screen.
    return:
        (tuple): (major version, minor version)
    """
    major_version = sys.version_info.major
    minor_version = sys.version_info.minor

    # display version
    if display_version:
        print(f"running on Python version {major_version}.{minor_version}")

    # check major version
    major_version_fail = False
    if min_major_version is not None:
        if not isinstance(min_major_version, (int, float)):
            raise TypeError(f"input parameter 'min_major_version is type "
                            f"({type(min_major_version)}), not int or float")
        if major_version < min_major_version:
            major_version_fail = True

    # check major version
    minor_version_fail = False
    if min_minor_version is not None:
        if not isinstance(min_minor_version, (int, float)):
            raise TypeError(f"input parameter 'min_minor_version is type "
                            f"({type(min_minor_version)}), not int or float")
        if minor_version < min_minor_version:
            minor_version_fail = True

    if major_version_fail or minor_version_fail:
        raise OSError(
            f"current python major version ({major_version}.{minor_version}) "
            f"is less than min python version required "
            f"({min_major_version}.{min_minor_version})")

    return (major_version, minor_version)


def dynamic_module_import(name, path=None, pkg=None, verbose=False):
    """
    Find and load python module.

    TODO: this module results in a resource warning within unittest:
    sys:1: ResourceWarning: unclosed <socket.socket fd=628,
    family=AddressFamily.AF_INET, type=SocketKind.SOCK_DGRAM, proto=0,
    laddr=('0.0.0.0', 64963)>

    inputs:
        name(str): module name
        path(str): file path (either relative or abs path),
                   if path is None then will import from installed packages
        pkg(str): package to add to path
        verbose(bool): debug flag
    returns:
        mod(module): module object
    """
    # add package to path
    if pkg is not None:
        pkg_path = get_parent_path(os.getcwd()) + "//" + pkg
        print(f"adding package '{pkg_path}' to path...")
        sys.path.insert(1, pkg_path)

    try:
        if path:
            # local file import from relative or abs path
            print(f"WARNING: attempting local import of {name} from "
                  f"working directory {os.getcwd()}...")
            if verbose:
                print(f"target dir contents={os.listdir(path)}")
            sys.path.insert(1, path)
            mod = importlib.import_module(name)
            if mod is None:
                raise ModuleNotFoundError(f"module '{name}' could not "
                                          f"be found at {path}")
        else:
            # installed package import
            spec = importlib.util.find_spec(name, path)
            if spec is None:
                raise ModuleNotFoundError(f"module '{name}' could "
                                          "not be found")
            mod = importlib.util.module_from_spec(spec)
            sys.modules[name] = mod
            spec.loader.exec_module(mod)
    except Exception as ex:
        util.log_msg(traceback.format_exc(),
                     mode=util.BOTH_LOG, func_name=1)
        util.log_msg("module load failed: " + name,
                     mode=util.BOTH_LOG, func_name=1)
        raise ex
    else:
        return mod


def get_parent_path(source_path, verbose=False):
    """
    Return the absolute path to the parent folder.

    inputs:
        source_path(str): source path
        verbose(bool): prints debug data.
    returns:
        (str): abs path to parent folder.
    """
    parent_path = os.path.abspath(os.path.join(
        source_path, os.pardir)).replace("\\", "/")
    if verbose:
        print(f"parent path={parent_path}")
    return parent_path


def get_package_version(module, element=None, verbose=False):
    """
    Get the package version for an installed package.

    inputs:
        module(obj): imported module
        element(str): element, if None will return the entire package
        verbose(bool): debug flag
    returns:
        (tuple) of ints if element is None, else returns int.
    """
    # cast lowercase if string
    if isinstance(element, str):
        element = element.lower()

    ver_tuple = tuple(map(int, module.__version__.split(".")))
    if verbose:
        print(f"{module} element {element} package version={ver_tuple}")
    if element is None:
        return_val = ver_tuple
    elif element in ["major", 0]:
        return_val = ver_tuple[0]
    elif element in ["minor", 1]:
        return_val = ver_tuple[1]
    elif element in ["patch", 2]:
        try:
            return_val = ver_tuple[2]
        except IndexError:
            return_val = 0
    else:
        raise AttributeError(f"{element} is not a valid choice "
                             "for element input")
    return return_val



"""
connection to 3m50 thermoststat on local net.
"""
# built-in imports
import datetime
import pprint
import socket
import time
import traceback
import urllib
from dns.exception import DNSException

# local imports
from thermostatsupervisor import mmm_config
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import thermostat_common as tc
from thermostatsupervisor import environment as env
from thermostatsupervisor import utilities as util

# radiotherm import
MMM_DEBUG = False  # debug uses local radiotherm repo instead of pkg
if MMM_DEBUG and not env.is_azure_environment():
    mod_path = "..//radiotherm"
    if env.is_interactive_environment():
        mod_path = "..\\" + mod_path
    radiotherm = env.dynamic_module_import("radiotherm",
                                           mod_path)
else:
    import radiotherm  # noqa E402, from path / site packages

SOCKET_TIMEOUT = 45  # http socket timeout override


class ThermostatClass(tc.ThermostatCommon):
    """3m50 thermostat zone functions."""

    def __init__(self, zone, verbose=True):
        """
        Constructor, connect to thermostat.

        inputs:
            zone(str):  zone of thermostat on local net.
                        mmm_config.metadata dict must have correct local IP
                        address for each zone.
            verbose(bool): debug flag.
        """
        # construct the superclass
        # tc.ThermostatCommonZone.__init__(self)
        # tc.ThermostatCommon.__init__(self)
        super().__init__()

        # set tstat type and debug flag
        self.thermostat_type = mmm_config.ALIAS
        self.verbose = verbose

        # configure zone info
        self.zone_name = int(zone)
        # use hard-coded IP address if provided, otherwise
        # use host dns lookup
        self.host_name = mmm_config.metadata[
            self.zone_name]["host_name"]
        # populate IP address from metadata dict.
        if "ip_address" in mmm_config.metadata[
                self.zone_name]:
            self.ip_address = mmm_config.metadata[
                self.zone_name]["ip_address"]
        else:
            # get IP address from DNS lookup on local net.
            ip_status, self.ip_address = util.is_host_on_local_net(
                self.host_name, verbose=True)
            if not ip_status:
                raise DNSException(
                    f"failed to resolve ip address for 3m thermostat "
                    f"'{self.host_name}'")
        self.device_id = self.get_target_zone_id()

    def get_target_zone_id(self) -> object:
        """
        Return the target zone ID from the
        zone number provided.

        inputs:
            None
        returns:
            (obj):  device object.
        """
        # verify thermostat exists on network
        try:
            self.device_id = radiotherm.get_thermostat(self.ip_address)
        except urllib.error.URLError as ex:
            raise RuntimeError(
                f"FATAL ERROR: 3m thermostat not found at ip address: "
                f"{self.ip_address}") from ex
        return self.device_id

    def print_all_thermostat_metadata(self, zone, debug=False):
        """
        Return initial meta data queried from thermostat.

        inputs:
            zone(int) zone number
            debug(bool): debug flag
        returns:
            None
        """
        # dump all meta data
        self.get_all_metadata(zone)

        # dump uiData in a readable format
        self.exec_print_all_thermostat_metadata(
            self.get_latestdata, [zone, debug])

    def get_meta_data_dict(self, zone) -> dict:
        """Build meta data dictionary from list of object attributes.

        inputs:
            zone(int): zone number
        returns:
            (dict) of meta data
        """
        util.log_msg(f"querying thermostat zone {zone} for meta data...",
                     mode=util.BOTH_LOG, func_name=1)
        attr_dict = {}
        ignore_fields = ['get', 'post', 'reboot', 'set_day_program']
        for attr in dir(self.device_id):
            if attr[0] != "_" and attr not in ignore_fields:
                key = attr
                try:
                    val = self.device_id.__getattribute__(key)
                except TypeError:
                    val = "<attribute is not readable>"
                except AttributeError as ex:
                    val = ex
                except socket.timeout:
                    val = "<socket timeout>"
                attr_dict[key] = val
        return attr_dict

    def get_all_metadata(self, zone) -> list:
        """
        Get all the current thermostat metadata.

        inputs:
            zone(int): zone number
        returns:
            (list) of thermostat attributes.
        """
        return self.get_meta_data_dict(zone)

    def get_metadata(self, zone, parameter=None) -> (dict, str):
        """
        Get the current thermostat metadata settings.

        inputs:
          zone(str or int): zone name
          parameter(str): target parameter, None = all settings
        returns:
          dict if parameter=None
          str if parameter != None
        """
        if parameter is None:
            return self.get_meta_data_dict(zone)
        else:
            return self.get_meta_data_dict(zone)[parameter]['raw']

    def get_latestdata(self, zone, debug=False) -> (dict, str):
        """
        Get the current thermostat latest data.

        inputs:
          zone(int): target zone
          debug(bool): debug flag
        returns:
          dict if parameter=None
          str if parameter != None
        """
        latest_data_dict = self.get_meta_data_dict(zone)
        if debug:
            util.log_msg(f"zone{zone} latestData: {latest_data_dict}",
                         mode=util.BOTH_LOG, func_name=1)
        return latest_data_dict

    def get_ui_data(self, zone) -> dict:
        """
        Get the latest thermostat ui data

        inputs:
          zone(int): target zone.
        returns:
          (dict)
        """
        return self.get_meta_data_dict(zone)

    def get_ui_data_param(self, zone, parameter) -> dict:
        """
        Get the latest thermostat ui data for one specific parameter.

        inputs:
          zone(int): target zone
          parameter(str): UI parameter
        returns:
          (various) value or data structure of interest
        """
        return self.get_meta_data_dict(zone)[parameter]['raw']


class ThermostatZone(tc.ThermostatCommonZone):
    """3m50 thermostat zone functions."""

    def __init__(self, Thermostat_obj, verbose=True):
        """
        Constructor, connect to thermostat.

        mmm_metadata dict above must have correct local IP address for each
            zone.

        inputs:
            Thermostat_obj(obj):  Thermostat class instance.
            verbose(bool): debug flag.
        """
        # construct the superclass
        super().__init__()

        # switch config for this thermostat
        self.system_switch_position[tc.ThermostatCommonZone.COOL_MODE] = 2
        self.system_switch_position[tc.ThermostatCommonZone.HEAT_MODE] = 1
        self.system_switch_position[tc.ThermostatCommonZone.OFF_MODE] = 0
        self.system_switch_position[tc.ThermostatCommonZone.AUTO_MODE] = 3

        # need verification
        self.system_switch_position[
            tc.ThermostatCommonZone.DRY_MODE] = util.BOGUS_INT
        self.system_switch_position[
            tc.ThermostatCommonZone.FAN_MODE] = util.BOGUS_INT

        # zone info
        self.verbose = verbose
        self.thermostat_type = mmm_config.ALIAS
        self.device_id = Thermostat_obj.device_id
        self.zone_name = Thermostat_obj.zone_name
        self.zone_name = self.get_zone_name()

        # runtime parameter defaults
        self.poll_time_sec = 10 * 60  # default to 10 minutes
        self.connection_time_sec = 8 * 60 * 60  # default to 8 hours

    def get_zone_name(self, zone=None):
        """
        Return the name associated with the zone number.

        inputs:
            zone(int): Zone number
        returns:
            (str) zone name
        """
        if zone is None:
            # pull from thermostat
            return self.device_id.name['raw']
        else:
            # override from config file
            return mmm_config.metadata[zone]["zone_name"]

    def get_display_temp(self) -> float:
        """
        Return DispTemperature.

        inputs:
            None
        returns:
            (float): display temp in degrees.
        """
        return float(self.device_id.temp['raw'])

    def get_display_humidity(self) -> (float, None):
        """
        Return Humidity.

        inputs:
            None
        returns:
            (float, None): humidity in %RH, return None if not supported.
        """
        return None  # not available

    def get_is_humidity_supported(self) -> bool:
        """
        Return humidity sensor status.

        inputs:
            None
        returns:
            (bool): True if humidity is supported.
        """
        return self.get_display_humidity() is not None

    def is_heat_mode(self) -> int:
        """
        Return the heat mode.

        inputs:
            None
        returns:
            (int): heat mode.
        """
        return int(self._get_tmode() ==
                   self.system_switch_position[self.HEAT_MODE])

    def is_cool_mode(self) -> int:
        """
        Return the cool mode.

        inputs:
            None
        returns:
            (int): 1=cool mode enabled, 0=disabled.
        """
        return int(self._get_tmode() ==
                   self.system_switch_position[self.COOL_MODE])

    def is_dry_mode(self) -> int:
        """
        Return the dry mode.

        inputs:
            None
        returns:
            (int): 1=dry mode enabled, 0=disabled.
        """
        return int(self._get_tmode() ==
                   self.system_switch_position[self.DRY_MODE])

    def is_auto_mode(self) -> int:
        """
        Return the auto mode.

        inputs:
            None
        returns:
            (int): 1=auto mode enabled, 0=disabled.
        """
        return int(self._get_tmode() ==
                   self.system_switch_position[self.AUTO_MODE])

    def _get_tmode(self, retries=1):
        """
        Get tmode from device, retry on Attribute error.
        """
        try:
            tmode = self.device_id.tmode['raw']
        except AttributeError as ex:
            if retries > 0:
                print(traceback.format_exc())
                print("WARNING: AttributeError while querying tstat.tmode, "
                      "retrying after brief delay...")
                time.sleep(10)
                tmode = self._get_tmode(retries-1)
            else:
                raise ex
        return tmode

    def is_fan_mode(self) -> int:
        """
        Return the fan mode.

        Fan mode is fan on for circulation but not applying heat or cooling.
        inputs:
            None
        returns:
            (int): 1=fan mode enabled, 0=disabled.
        """
        return int(self.device_id.fmode['raw'] == 2 and self.is_off_mode())

    def is_off_mode(self) -> int:
        """
        Return the off mode.

        inputs:
            None
        returns:
            (int): off mode, 1=enabled, 0=disabled.
        """
        return int(self._get_tmode() ==
                   self.system_switch_position[self.OFF_MODE] and
                   not self.device_id.fmode['raw'] == 2)

    def is_heating(self):
        """Return 1 if heating relay is active, else 0."""
        return int(self.is_heat_mode() and self.is_power_on() and
                   self.get_heat_setpoint_raw() > self.get_display_temp())

    def is_cooling(self):
        """Return 1 if cooling relay is active, else 0."""
        return int(self.is_cool_mode() and self.is_power_on() and
                   self.get_cool_setpoint_raw() < self.get_display_temp())

    def is_drying(self):
        """Return 1 if drying relay is active, else 0."""
        return 0  # not applicable

    def is_auto(self):
        """Return 1 if auto relay is active, else 0."""
        return 0  # not applicable

    def is_fanning(self):
        """Return 1 if fan relay is active, else 0."""
        return int(self.is_fan_on() and self.is_power_on())

    def is_power_on(self):
        """Return 1 if power relay is active, else 0."""
        return int(self.device_id.power['raw'] > 0)

    def is_fan_on(self):
        """Return 1 if fan relay is active, else 0."""
        return self.device_id.fstate['raw']

    def is_defrosting(self):
        """Return 1 if defrosting is active, else 0."""
        return 0  # not applicable

    def is_standby(self):
        """Return 1 if standby is active, else 0."""
        return 0  # not applicable

    def get_setpoint_list(self, sp_dict, day) -> list:
        """
        Return list of 4 setpoints for the day.

        inputs:
            sp_dict(dict): setpoint dictionary
            day(int): day of the week, 0=Monday
        returns:
            list of 8 elements, representing 4 pairs of elapsed minutes
            and setpoint
        """
        sp_lst = sp_dict['raw'][str(day)]
        if len(sp_lst) != 8:
            raise ValueError(f"setpoint list is not 8 elements long: {sp_lst}")
        return sp_dict['raw'][str(day)]

    def get_previous_days_setpoint(self, sp_dict) -> int:
        """
        Return last setpoint from day before today.

        inputs:
            sp_dict(dict): setpoint dictionary
        returns:
            (int)last temperature set the day before.
        """
        today = datetime.datetime.today().weekday()
        if today == 0:
            yesterday = 6
        else:
            yesterday = today - 1
        return self.get_setpoint_list(sp_dict, yesterday)[-1]

    def get_schedule_setpoint(self, sp_dict) -> int:
        """
        Return current setpoint.

        inputs:
            sp_dict(dict): setpoint dictionary
        returns:
            (int) last temperature set the day before.
        """
        current_sp = self.get_previous_days_setpoint(sp_dict)
        now = datetime.datetime.now()
        minutes_since_midnight = (
            now - now.replace(hour=0, minute=0, second=0, microsecond=0)
        ).total_seconds() / 60
        todays_setpoint_lst = self.get_setpoint_list(sp_dict,
                                                     datetime.datetime.today(
                                                     ).weekday())
        for idx in [0, 2, 4, 6]:
            if todays_setpoint_lst[idx] > minutes_since_midnight:
                return current_sp
            else:
                # store setpoint corresponding to this time
                current_sp = todays_setpoint_lst[idx + 1]
        return current_sp

    def get_heat_setpoint(self) -> float:
        """
        Return the current heat setpoint.

        t_heat field is only present if in heat mode.
        inputs:
            None
        returns:
            (float): current heat set point in degrees.
        """
        result = self.device_id.t_heat['raw']
        if not isinstance(result, float):
            raise TypeError(f"heat set point is type {type(result)}, "
                            f"should be float")
        return result

    def get_heat_setpoint_raw(self) -> float:
        """
        Return the current raw heat setpoint.

        t_heat field is only present if in heat mode.
        inputs:
            None
        returns:
            (float): current raw heat set point in degrees.
        """
        result = self.get_heat_setpoint()
        if not isinstance(result, float):
            raise TypeError(
                f"heat set point raw is type {type(result)}, should be float")
        return result

    def get_cool_setpoint(self) -> int:
        """
        Return the current cool setpoint.

        t_cool field is only present if in cool mode.
        inputs:
            None
        returns:
            (int): current cool set point in degrees.
        """
        result = self.device_id.t_cool['raw']
        if not isinstance(result, (int, float)):
            raise TypeError(
                f"cool set point is type {type(result)}, should be "
                f"(int, float)")
        return result

    def get_cool_setpoint_raw(self) -> int:
        """
        Return the current raw cool setpoint.

        t_cool field is only present if in cool mode.
        inputs:
            None
        returns:
            (int): current raw cool set point in degrees.
        """
        result = self.get_cool_setpoint()
        if not isinstance(result, (int, float)):
            raise TypeError(
                f"cool setpoint raw is type {type(result)}, should be "
                f"(int, float)")
        return result

    def get_schedule_program_heat(self) -> dict:
        """
        Return the scheduled heat program, times and settings.

        inputs:
            None
        returns:
            (dict): scheduled heat set points and times in degrees.
        """
        result = self.device_id.program_heat['raw']
        if not isinstance(result, dict):
            raise TypeError(
                f"heat program schedule set point is type {type(result)},"
                f" should be dict")
        return result

    def get_schedule_heat_sp(self) -> int:
        """
        Return the scheduled heat setpoint.

        inputs:
            None
        returns:
            (int): current scheduled heat set point in degrees.
        """
        result = self.get_schedule_setpoint(self.device_id.program_heat)
        if not isinstance(result, int):
            raise TypeError(
                f"schedule heat set point is type {type(result)}, "
                f"should be int")
        return result

    def get_schedule_program_cool(self) -> dict:
        """
        Return the sechduled cool setpoint.

        inputs:
            None
        returns:
            (dict): current scheduled cool set point in degrees.
        """
        result = self.device_id.program_cool['raw']
        if not isinstance(result, dict):
            raise TypeError(
                f"schedule program cool set point is type {type(result)},"
                f" should be dict")
        return result

    def get_schedule_cool_sp(self) -> int:
        """
        Return the sechduled cool setpoint.

        inputs:
            None
        returns:
            (int): current schedule cool set point in degrees.
        """
        result = self.get_schedule_setpoint(self.device_id.program_cool)
        if not isinstance(result, int):
            raise TypeError(
                f"schedule cool set point is type {type(result)}, "
                f"should be int")
        return result

    def get_is_invacation_hold_mode(self) -> bool:
        """
        Return the in vacation hold status.

        inputs:
            None
        returns:
            (int): 0=Disabled, 1=Enabled
        """
        result = bool(self.device_id.hold['raw'])
        if not isinstance(result, bool):
            raise TypeError(
                f"is_invacation_hold_mode is type {type(result)}, "
                f"should be bool")
        return result

    def get_vacation_hold(self) -> bool:
        """
        Return the Hold setting.

        inputs:
            None
        returns:
            (int): 0=Disabled, 1=Enabled
        """
        result = bool(self.device_id.override['raw'])
        if not isinstance(result, bool):
            raise TypeError(
                f"get_vacation_hold_mode is type {type(result)}, "
                f"should be bool")
        return result

    def get_vacation_hold_until_time(self) -> int:
        """
        Return the 'VacationHoldUntilTime'.

        inputs:
            None
        returns:
            (int): vacation hold time in minutes.
         """
        return -1  # not implemented

    def get_temporary_hold_until_time(self) -> int:
        """
        Return the 'TemporaryHoldUntilTime'.

        inputs:
            None
        returns:
            (int): temporary hold time in minutes.
         """
        if self.is_heat_mode() == 1:
            sp_dict = self.device_id.program_heat
        elif self.is_cool_mode() == 1:
            sp_dict = self.device_id.program_cool
        else:
            # off mode, use dummy dict.
            sp_dict = {'raw': {'0': [0] * 8, '1': [0] * 8, '2': [0] * 8,
                               '3': [0] * 8, '4': [0] * 8, '5': [0] * 8,
                               '6': [0] * 8}}
            # raise ValueError("unknown heat/cool mode")
        now = datetime.datetime.now()
        minutes_since_midnight = (
            now - now.replace(hour=0, minute=0, second=0,
                              microsecond=0)).total_seconds() / 60
        todays_setpoint_lst = self.get_setpoint_list(sp_dict,
                                                     datetime.datetime.today(
                                                     ).weekday())
        for idx in [0, 2, 4, 6]:
            if todays_setpoint_lst[idx] > minutes_since_midnight:
                time_delta = (todays_setpoint_lst[idx] -
                              minutes_since_midnight)
                break  # found next timer
            else:
                time_delta = 24 * 60 - minutes_since_midnight

        return int(time_delta)

    def get_setpoint_change_allowed(self) -> bool:
        """
        Return the 'SetpointChangeAllowed' setting.

        inputs:
            None
        returns:
            (bool): 'SetpointChangeAllowed' will be True in heating mode,
            False in OFF mode (assume True in cooling mode too)
        """
        return False  # not implemented

    def get_system_switch_position(self) -> int:
        """ Return the thermostat mode.

        inputs:
            None
        returns:
            (int): thermostat mode, refer to tc.system_swtich position
            for details.
        """
        result = self._get_tmode()
        if not isinstance(result, int):
            raise TypeError(
                f"get_system_switch_position is type {type(result)}, "
                f"should be int")
        return result

    def set_heat_setpoint(self, temp: int) -> None:
        """
        Set a new heat setpoint.

        This will also attempt to turn the thermostat to 'Heat'
        inputs:
            temp(int): desired temperature in degrees.
        returns:
            None
        """
        self.device_id.t_heat = temp

    def set_cool_setpoint(self, temp: int) -> None:
        """
        Set a new cool setpoint.

        This will also attempt to turn the thermostat to 'Cool'
        inputs:
            temp(int): desired temperature in degrees.
        returns:
            None
        """
        self.device_id.t_cool = temp

    def report_heating_parameters(self, switch_position=None):
        """
        Display critical thermostat settings and reading to the screen.

        inputs:
            switch_position(int): switch position override, used for testing.
        returns:
            None
        """
        # current temp as measured by thermostat
        util.log_msg(
            f"display temp="
            f"{util.temp_value_with_units(self.get_display_temp())}",
            mode=util.BOTH_LOG,
            func_name=1)

        # get switch position
        if switch_position is None:
            switch_position = self.get_system_switch_position()

        # heating status
        if switch_position == \
                self.system_switch_position[self.HEAT_MODE]:
            util.log_msg(f"heat mode={self.is_heat_mode()}",
                         mode=util.BOTH_LOG)
            util.log_msg(
                f"heat setpoint={self.get_heat_setpoint_raw()}",
                mode=util.BOTH_LOG)
            util.log_msg(
                f"schedule heat sp={self.get_schedule_heat_sp()}",
                mode=util.BOTH_LOG)

        # cooling status
        if switch_position == \
                self.system_switch_position[self.COOL_MODE]:
            util.log_msg(f"cool mode={self.is_cool_mode()}",
                         mode=util.BOTH_LOG)
            util.log_msg(
                f"cool setpoint={self.get_cool_setpoint_raw()}",
                mode=util.BOTH_LOG)
            util.log_msg(
                f"schedule cool sp={self.get_schedule_cool_sp()}",
                mode=util.BOTH_LOG)

        # hold settings
        util.log_msg(
            f"is in vacation hold mode={self.get_is_invacation_hold_mode()}",
            mode=util.BOTH_LOG)
        util.log_msg(f"vacation hold={self.get_vacation_hold()}",
                     mode=util.BOTH_LOG)
        util.log_msg(
            f"vacation hold until time={self.get_vacation_hold_until_time()}",
            mode=util.BOTH_LOG)
        util.log_msg(
            f"temporary hold until time="
            f"{self.get_temporary_hold_until_time()}",
            mode=util.BOTH_LOG)


# monkeypatch radiotherm.thermostat.Thermostat __init__ method with longer
# socket.timeout delay, based on radiotherm version 2.1
def __init__(self, host, timeout=SOCKET_TIMEOUT):  # changed from 4 (default)
    self.host = host
    self.timeout = timeout


radiotherm.thermostat.Thermostat.__init__ = __init__
# end of monkeypatch radiotherm.thermostat.Thermostst __init__

if __name__ == "__main__":

    # verify environment
    env.get_python_version()

    # get zone override
    api.uip = api.UserInputs(argv_list=None,
                             thermostat_type=mmm_config.ALIAS)
    zone_number = api.uip.get_user_inputs(api.uip.zone_name,
                                          api.input_flds.zone)

    _, Zone = tc.thermostat_basic_checkout(
        mmm_config.ALIAS,
        zone_number,
        ThermostatClass, ThermostatZone)

    tc.thermostat_get_all_zone_temps(
        mmm_config.ALIAS,
        mmm_config.supported_configs["zones"],
        ThermostatClass,
        ThermostatZone)

    # measure thermostat response time
    MEASUREMENTS = 60  # running higher than normal count here because
    # intermittent failures have been observed.
    print(f"Thermostat response times for {MEASUREMENTS} measurements...")
    meas_data = Zone.measure_thermostat_response_time(MEASUREMENTS)
    ppp = pprint.PrettyPrinter(indent=4)
    ppp.pprint(meas_data)



"""
Flask server for displaying supervisor output on web page.
"""
# built-in libraries
import html
from subprocess import Popen, PIPE, STDOUT, DEVNULL
import sys
import webbrowser

# third party imports
from flask import Flask, Response
from flask_wtf.csrf import CSRFProtect

# local imports
from thermostatsupervisor import environment as env
from thermostatsupervisor import flask_generic as flg
from thermostatsupervisor import supervise as sup
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import utilities as util

# flask server
if util.is_windows_environment():
    # win server from Eclipse IDE:
    #     loopback will work to itself but not remote clients
    #     local IP works both itself and to remote Linux client.
    # win server from command line:
    #
    flask_ip_address = env.get_local_ip()
else:
    # Linux server from Thoney IDE: must update Thonny to run from root
    #   page opens on both loopback Linux and remote Win client, but
    #       no data loads.
    # flask_ip_address = '127.0.0.1'  # almost works from Linux client
    flask_ip_address = '0.0.0.0'
    # on Linux both methds are returning correct page header, but no data
FLASK_PORT = 5001  # note: ports below 1024 require root access on Linux
FLASK_USE_HTTPS = False  # HTTPS requires a cert to be installed.
if FLASK_USE_HTTPS:
    FLASK_SSL_CERT = 'adhoc'  # adhoc
    flask_kwargs = {'ssl_context': FLASK_SSL_CERT}
    FLASK_URL_PREFIX = "https://"
else:
    FLASK_SSL_CERT = None  # adhoc
    flask_kwargs = {}
    FLASK_URL_PREFIX = "http://"
flask_url = FLASK_URL_PREFIX + flask_ip_address + ':' + str(FLASK_PORT)

argv = []  # supervisor runtime args list


def create_app():
    """Create the flask object."""
    app_ = Flask(__name__)
    # api = Api(app)

    # api.add_resource(Controller, "/")
    return app_


# create the flask app
app = create_app()
csrf = CSRFProtect(app)  # enable CSRF protection
ip_ban = flg.initialize_ipban(app)  # hacker blacklisting agent
flg.set_flask_cookie_config(app)
flg.print_flask_config(app)


@app.route('/favicon.ico')
def favicon():
    """Faviocon displayed in browser tab."""
    return app.send_static_file('honeywell.ico')


@app.route('/')
def index():
    """index route"""
    def run_supervise():
        sup.argv = argv  # pass runtime overrides to supervise
        api.uip = api.UserInputs(argv)
        thermostat_type = api.uip.get_user_inputs(
            api.uip.zone_name,
            api.input_flds.thermostat_type)
        zone = api.uip.get_user_inputs(api.uip.zone_name,
                                       api.input_flds.zone)
        measurement_cnt = api.uip.get_user_inputs(
            api.uip.zone_name,
            api.input_flds.measurements)
        title = (f"{thermostat_type} thermostat zone {zone}, "
                 f"{measurement_cnt} measurements")
        yield f"<!doctype html><title>{title}</title>"

        # runtime variabless
        executable = "python"
        dont_buffer = "-u"  # option to not buffer results
        run_module = "-m"  # option to reference package
        script = "thermostatsupervisor.supervise"
        if argv:
            # argv list override for unit testing
            arg_list = [executable, dont_buffer, run_module, script] + argv[1:]
        elif len(sys.argv) > 1:
            arg_list = ([executable, dont_buffer, run_module, script] +
                        sys.argv[1:])
        else:
            arg_list = [executable, dont_buffer, run_module, script]
        with Popen(arg_list, stdin=DEVNULL, stdout=PIPE, stderr=STDOUT,
                   bufsize=1, universal_newlines=True, shell=True) as p_out:
            for i, line in enumerate(p_out.stdout):
                print(f"DEBUG: line {i}: {line}", file=sys.stderr)
                yield "<code>{}</code>".format(html.escape(line.rstrip("\n")))
                yield "<br>\n"
    return Response(run_supervise(), mimetype='text/html')


if __name__ == '__main__':
    # show the page in browser
    webbrowser.open(flask_url, new=2)
    flg.schedule_ipban_block_list_report(ip_ban, debug_mode=False)
    app.run(host=flask_ip_address,
            port=FLASK_PORT,
            debug=False,  # True causes 2 tabs to open, enables auto-reload
            threaded=True,  # threaded=True may speed up rendering on web page
            ssl_context=FLASK_SSL_CERT)



"""
Common Thermostat Class
"""
# built-ins
import datetime
import operator
import pprint
import statistics
import time
import traceback

# local imports
from thermostatsupervisor import email_notification as eml
from thermostatsupervisor import thermostat_api as api
from thermostatsupervisor import utilities as util


DEGREE_SIGN = "\N{DEGREE SIGN}"

connection_ok = True  # global flag for connection OK.
connection_fail_cnt = 0  # count number of connection failures
max_connection_fail_cnt = 9  # max number of connection failures


class ThermostatCommon():
    """Class methods common to all thermostat objects."""

    def __init__(self, *_, **__):
        self.verbose = True
        self.thermostat_type = "unknown"  # placeholder
        self.zone_number = util.BOGUS_INT  # placeholder
        self.device_id = util.BOGUS_INT  # placeholder
        self.ip_address = None  # placeholder
        global connection_ok  # noqa W603
        connection_ok = True

    def print_all_thermostat_metadata(self, zone, debug=False):  # noqa R0201
        """
        Print initial meta data queried from thermostat for specified zone.

        inputs:
            zone(int): zone number
            debug(bool): debug flag
        returns:
            None
        """
        del debug  # placeholder function
        util.log_msg(
            f"WARNING: print_all_thermostat_metatdata({zone}) not yet "
            f"implemented for this thermostat type\n",
            mode=util.BOTH_LOG,
            func_name=1)

    def exec_print_all_thermostat_metadata(self, func, args):
        """
        Print all metadata to screen.

        inputs:
            func(obj): function get metadata.
            args(list): argument list
        returns:
            (dict): return data
        """
        # dump metadata in a readable format
        return_data = func(*args)
        pprint_obj = pprint.PrettyPrinter(indent=4)
        print("\n")
        util.log_msg("raw thermostat meta data:",
                     mode=util.BOTH_LOG, func_name=1)
        pprint_obj.pprint(return_data)
        return return_data


class ThermostatCommonZone():
    """Class methods common to all thermostat zones."""

    # supported thermostat modes and label text
    OFF_MODE = "OFF_MODE"
    HEAT_MODE = "HEAT_MODE"
    COOL_MODE = "COOL_MODE"
    AUTO_MODE = "AUTO_MODE"
    DRY_MODE = "DRY_MODE"
    FAN_MODE = "FAN_MODE"
    UNKNOWN_MODE = "UNKNOWN_MODE"  # bypass set mode or unable to detect

    # modes where heat is applied
    heat_modes = [HEAT_MODE, AUTO_MODE]

    # modes where cooling is applied
    cool_modes = [COOL_MODE, DRY_MODE, AUTO_MODE]

    # modes in which setpoints apply
    controlled_modes = [HEAT_MODE, AUTO_MODE, COOL_MODE]

    system_switch_position = {
        # placeholder, will be tstat-specific
        UNKNOWN_MODE: util.BOGUS_INT,
        HEAT_MODE: util.BOGUS_INT - 1,
        COOL_MODE: util.BOGUS_INT - 2,
        AUTO_MODE: util.BOGUS_INT - 3,
        DRY_MODE: util.BOGUS_INT - 4,
        FAN_MODE: util.BOGUS_INT - 5,
        OFF_MODE: util.BOGUS_INT - 6,
    }
    max_scheduled_heat_allowed = 74  # warn if scheduled heat value exceeds.
    min_scheduled_cool_allowed = 68  # warn if scheduled cool value exceeds.
    tolerance_degrees_default = 2  # allowed override vs. the scheduled value.

    def __init__(self, *_, **__):
        self.verbose = True
        self.thermostat_type = "unknown"  # placeholder
        self.zone_number = util.BOGUS_INT  # placeholder
        self.zone_name = None  # placeholder
        self.device_id = util.BOGUS_INT  # placeholder
        self.poll_time_sec = util.BOGUS_INT  # placeholder
        self.session_start_time_sec = util.BOGUS_INT  # placeholder
        self.connection_time_sec = util.BOGUS_INT  # placeholder
        self.target_mode = "OFF_MODE"  # placeholder
        self.flag_all_deviations = False  #
        self.revert_deviations = True  # Revert deviation
        self.revert_all_deviations = False  # Revert only energy-wasting events
        self.temperature_is_deviated = False  # temp deviated from schedule
        self.display_temp = None  # current temperature in deg F
        self.display_humidity = None  # current humidity in %RH
        self.humidity_is_available = False  # humidity supported flag
        self.hold_mode = False  # True = not following schedule
        self.hold_temporary = False
        self.zone_info = {}  # dict containing zone data
        self.last_fetch_time = None  # last fetch of zone_info

        # abstraction vars and funcs, defined in query_thermostat_zone
        self.current_mode = None  # str representing mode
        self.current_setpoint = util.BOGUS_INT  # current setpoint
        self.schedule_setpoint = util.BOGUS_INT  # current scheduled setpoint
        self.tolerance_sign = 1  # +1 for heat, -1 for cool
        self.operator = operator.ne  # operator for deviation check
        self.tolerance_degrees = self.tolerance_degrees_default
        self.global_limit = util.BOGUS_INT  # global temp limit
        self.global_operator = operator.ne  # oper for global temp deviation
        self.revert_setpoint_func = self.function_not_supported
        self.get_setpoint_func = self.function_not_supported

    def query_thermostat_zone(self):
        """Return the current mode and set mode-specific parameters."""

        # current temperature
        try:
            self.display_temp = self.validate_numeric(self.get_display_temp(),
                                                      "get_display_temp")
        except TypeError:
            util.log_msg(traceback.format_exc(),
                         mode=util.BOTH_LOG, func_name=1)
            self.display_temp = None

        # current humidity
        self.display_humidity = self.get_display_humidity()
        self.humidity_is_available = self.get_is_humidity_supported()

        # mode-specific parameters
        if self.is_heat_mode():
            self.current_mode = self.HEAT_MODE
            self.current_setpoint = int(self.get_heat_setpoint_raw())
            self.schedule_setpoint = int(self.get_schedule_heat_sp())
            self.tolerance_sign = 1
            if self.flag_all_deviations:
                self.operator = operator.ne
                self.tolerance_degrees = 0  # disable tolerance
            else:
                self.operator = operator.gt
            self.global_limit = self.max_scheduled_heat_allowed
            self.global_operator = operator.gt
            self.revert_setpoint_func = self.set_heat_setpoint
            self.get_setpoint_func = self.get_heat_setpoint_raw
        elif self.is_cool_mode():
            self.current_mode = self.COOL_MODE
            self.current_setpoint = int(self.get_cool_setpoint_raw())
            self.schedule_setpoint = int(self.get_schedule_cool_sp())
            self.tolerance_sign = -1
            if self.flag_all_deviations:
                self.operator = operator.ne
                self.tolerance_degrees = 0  # disable tolerance
            else:
                self.operator = operator.lt
            self.global_limit = self.min_scheduled_cool_allowed
            self.global_operator = operator.lt
            self.revert_setpoint_func = self.set_cool_setpoint
            self.get_setpoint_func = self.get_cool_setpoint_raw
        elif self.is_dry_mode():
            self.current_mode = self.DRY_MODE
            self.current_setpoint = int(self.get_cool_setpoint_raw())
            self.schedule_setpoint = int(self.get_schedule_cool_sp())
            self.tolerance_sign = -1
            if self.flag_all_deviations:
                self.operator = operator.ne
                self.tolerance_degrees = 0  # disable tolerance
            else:
                self.operator = operator.lt
            self.global_limit = self.min_scheduled_cool_allowed
            self.global_operator = operator.lt
            self.revert_setpoint_func = self.function_not_supported
            self.get_setpoint_func = self.function_not_supported
        elif self.is_auto_mode():
            self.current_mode = self.AUTO_MODE
            self.current_setpoint = util.BOGUS_INT
            self.schedule_setpoint = util.BOGUS_INT
            self.tolerance_sign = 1
            self.operator = operator.ne
            self.global_limit = util.BOGUS_INT
            self.global_operator = operator.ne
            self.revert_setpoint_func = self.function_not_supported
            self.get_setpoint_func = self.function_not_supported
        elif self.is_fan_mode():
            self.current_mode = self.FAN_MODE
            self.current_setpoint = util.BOGUS_INT
            self.schedule_setpoint = util.BOGUS_INT
            self.tolerance_sign = 1
            self.operator = operator.ne
            self.global_limit = util.BOGUS_INT
            self.global_operator = operator.ne
            self.revert_setpoint_func = self.function_not_supported
            self.get_setpoint_func = self.function_not_supported
        elif self.is_off_mode():
            self.current_mode = self.OFF_MODE
            self.current_setpoint = util.BOGUS_INT
            self.schedule_setpoint = util.BOGUS_INT
            self.tolerance_sign = 1
            self.operator = operator.ne
            self.global_limit = util.BOGUS_INT
            self.global_operator = operator.ne
            self.revert_setpoint_func = self.function_not_supported
            self.get_setpoint_func = self.function_not_supported
        else:
            print(f"DEBUG: zone info: {self.zone_info}")
            raise ValueError("unknown thermostat mode")

        self.temperature_is_deviated = self.is_temp_deviated_from_schedule()

    def is_temp_deviated_from_schedule(self):
        """
        Return True if temperature is deviated from current schedule.

        inputs:
            None:
        returns:
            (bool): True if temp is deviated from schedule.
        """
        return self.operator(self.current_setpoint, self.schedule_setpoint
                             + self.tolerance_sign
                             * self.tolerance_degrees)

    def get_current_mode(self, session_count, poll_count, print_status=True,
                         flag_all_deviations=False):
        """
        Determine whether thermostat is following schedule or if it has been
        deviated from schedule.

        inputs:
            session_count(int): session number (connection #) for reporting
            poll_count(int): poll number for reporting
            print_status(bool):  True to print status line
            flag_all_deviations(bool):  True: flag all deviations
                                        False(default): only flag energy
                                                        consuming deviations,
                                                        e.g. heat setpoint
                                                        above schedule,
                                                        cool setpoint
                                                        below schedule
        returns:
            dictionary of heat/cool mode status, deviation status,
            and hold status
        """

        return_buffer = {
            "heat_mode": util.BOGUS_BOOL,  # in heating mode
            "cool_mode": util.BOGUS_BOOL,  # in cooling mode
            "heat_deviation": util.BOGUS_BOOL,  # True: heat is deviated above
            "cool_deviation": util.BOGUS_BOOL,  # True: cool is deviated below
            "hold_mode": util.BOGUS_BOOL,  # True if hold is enabled
            "status_msg": "",  # status message
        }

        self.flag_all_deviations = flag_all_deviations
        self.query_thermostat_zone()

        # warning email if set point is outside global limit
        self.warn_if_outside_global_limit(self.current_setpoint,
                                          self.global_limit,
                                          self.global_operator,
                                          self.current_mode)

        if self.is_temp_deviated_from_schedule() and self.is_controlled_mode():
            status_msg = (f"[{self.current_mode.upper()} deviation] act temp="
                          f"{util.temp_value_with_units(self.display_temp)}")
        else:
            status_msg = (f"[following schedule] act temp="
                          f"{util.temp_value_with_units(self.display_temp)}")

        # add humidity if available
        if self.humidity_is_available:
            status_msg += \
                (f", act humidity="
                 f"{util.humidity_value_with_units(self.display_humidity)}")

        # add hold information
        if self.is_temp_deviated_from_schedule() and self.is_controlled_mode():
            self.hold_mode = True  # True = not following schedule
            self.hold_temporary = self.get_temporary_hold_until_time() > 0
            status_msg += (
                f" ({['persistent', 'temporary'][self.hold_temporary]})")
        else:
            self.hold_mode = False
            self.hold_temporary = False

        # add setpoints if in heat or cool mode
        if self.is_heat_mode() or self.is_cool_mode():
            status_msg += (
                f", set point="
                f"{util.temp_value_with_units(self.schedule_setpoint)}, "
                f"tolerance="
                f"{util.temp_value_with_units(self.tolerance_degrees)}, "
                f"override="
                f"{util.temp_value_with_units(self.current_setpoint)}")

        date_str = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        full_status_msg = (
            f"{date_str}: "
            f"(tstat:{self.thermostat_type}, zone:{self.zone_name}, "
            f"session:{session_count}, poll:{poll_count}) "
            f"{self.current_mode.upper()} {status_msg}")
        if print_status:
            util.log_msg(full_status_msg, mode=util.BOTH_LOG)

        self.store_current_mode()

        # return status
        return_buffer["heat_mode"] = self.is_heat_mode()
        return_buffer["cool_mode"] = self.is_cool_mode()
        return_buffer["heat_deviation"] = self.is_heat_deviation()
        return_buffer["cool_deviation"] = self.is_cool_deviation()
        return_buffer["hold_mode"] = self.hold_mode
        return_buffer["status_msg"] = full_status_msg
        return return_buffer

    def set_mode(self, target_mode):
        """
        Set the thermostat mode.

        inputs:
            target_mode(str): target mode, refer to supported_configs["modes"]
        returns:
            True if successful, else False
        """
        print(f"DEBUG in set_mode, target_mode={target_mode}, doing nothing")
        return False

    def store_current_mode(self):
        """Save the current mode to cache."""
        if self.is_heat_mode():
            self.current_mode = self.HEAT_MODE
        elif self.is_cool_mode():
            self.current_mode = self.COOL_MODE
        elif self.is_dry_mode():
            self.current_mode = self.DRY_MODE
        elif self.is_auto_mode():
            self.current_mode = self.AUTO_MODE
        elif self.is_fan_mode():
            self.current_mode = self.FAN_MODE
        elif self.is_off_mode():
            self.current_mode = self.OFF_MODE
        else:
            raise ValueError("unknown thermostat mode")

    def validate_numeric(self, input_val, parameter_name):
        """
        Validate value returned is numeric, otherwise raise exception.

        inputs:
            input_val: input value of unknown type.
            parameter_name(str): parameter name.
        returns:
            (int, float): pass thru value if numeric, else raise exception.
        """
        if not isinstance(input_val, (int, float)):
            raise TypeError(
                f"value returned for parameter '{parameter_name}' is "
                f"type {type(input_val)}, expected int or float")
        return input_val

    def warn_if_outside_global_limit(self, setpoint, limit_value, oper, label):
        """
        Send warning email if setpoint is outside of global limits.

        inputs:
            setpoint(int): setpoint value.
            limit_value(int): the limit value
            oper(operator):  the operator, either operator.gt or operator.lt
            label(str): label for warning message denoting the mode
        returns:
            (bool): result of check
        """
        if oper == operator.gt:  # pylint: disable=W0143
            level = "above max"
        else:
            level = "below min"
        if oper(setpoint, limit_value):
            msg = (f"{self.thermostat_type} zone {self.zone_name}: "
                   f"scheduled {label.upper()} set point "
                   f"({util.temp_value_with_units(setpoint)}) is {level} "
                   f"limit ({util.temp_value_with_units(limit_value)})")
            util.log_msg(f"WARNING: {msg}", mode=util.BOTH_LOG)
            eml.send_email_alert(
                subject=msg,
                body=f"{util.get_function_name()}: {msg}")
            return True
        else:
            return False

    def is_heat_mode(self):
        """Return True if in heat mode."""
        return (self.get_system_switch_position() ==
                self.system_switch_position[self.HEAT_MODE])

    def is_cool_mode(self):
        """Return True if in cool mode."""
        return (self.get_system_switch_position() ==
                self.system_switch_position[self.COOL_MODE])

    def is_dry_mode(self):
        """Return True if in dry mode."""
        return (self.get_system_switch_position() ==
                self.system_switch_position[self.DRY_MODE])

    def is_auto_mode(self):
        """Return True if in auto mode."""
        return (self.get_system_switch_position() ==
                self.system_switch_position[self.AUTO_MODE])

    def is_fan_mode(self):
        """Return 1 if fan mode enabled, else 0."""
        return (self.get_system_switch_position() ==
                self.system_switch_position[self.FAN_MODE])

    def is_off_mode(self):
        """Return 1 if fan mode enabled, else 0."""
        return (self.get_system_switch_position() ==
                self.system_switch_position[self.OFF_MODE])

    def is_controlled_mode(self):
        """Return True if mode is being controlled."""
        return self.current_mode in self.controlled_modes

    def is_heating(self):  # noqa R0201
        """Return 1 if heating relay is active, else 0."""
        return util.BOGUS_INT

    def is_cooling(self):  # noqa R0201
        """Return 1 if cooling relay is active, else 0."""
        return util.BOGUS_INT

    def is_drying(self):  # noqa R0201
        """Return 1 if drying relay is active, else 0."""
        return util.BOGUS_INT

    def is_auto(self):  # noqa R0201
        """Return 1 if auto relay is active, else 0."""
        return util.BOGUS_INT

    def is_fanning(self):  # noqa R0201
        """Return 1 if fan relay is active, else 0."""
        return util.BOGUS_INT

    def set_heat_setpoint(self, temp: int) -> None:  # noqa R0201
        """
        Set a new heat setpoint.

        This will also attempt to turn the thermostat to 'Cool'
        inputs:
            temp(int): desired temperature in deg F.
        returns:
            None
        """
        del temp
        util.log_msg(
            f"WARNING: {util.get_function_name()}: function is not yet "
            f"implemented on this thermostat, doing nothing",
            mode=util.BOTH_LOG,
            func_name=1)

    def set_cool_setpoint(self, temp: int) -> None:  # noqa R0201
        """
        Set a new cool setpoint.

        This will also attempt to turn the thermostat to 'Cool'
        inputs:
            temp(int): desired temperature in deg F.
        returns:
            None
        """
        del temp
        util.log_msg(
            f"WARNING: {util.get_function_name()}: function is not yet "
            f"implemented on this thermostat, doing nothing",
            mode=util.BOTH_LOG,
            func_name=1)

    def is_heat_deviation(self):
        """
        Return True if heat is deviated.

        inputs:
            None
        returns:
            (bool): True if deviation exists.
        """
        return self.is_heat_mode() and self.is_temp_deviated_from_schedule()

    def is_cool_deviation(self):
        """
        Return True if cool is deviated.

        inputs:
            None
        returns:
            (bool): True if deviation exists.
        """
        return self.is_cool_mode() and self.is_temp_deviated_from_schedule()

    # Thermostat-specific methods will be overloaded
    def get_display_temp(self) -> float:  # noqa R0201
        """Return the displayed temperature."""
        return float(util.BOGUS_INT)  # placeholder

    def get_display_humidity(self) -> float:  # noqa R0201
        """Return the displayed humidity."""
        return float(util.BOGUS_INT)  # placeholder

    def get_is_humidity_supported(self) -> bool:  # noqa R0201
        """Return humidity sensor status."""
        return util.BOGUS_BOOL  # placeholder

    def get_system_switch_position(self) -> int:  # noqa R0201
        """Return the 'SystemSwitchPosition'
            'SystemSwitchPosition' = 1 for heat, 2 for off
        """
        return util.BOGUS_INT  # placeholder

    def get_heat_setpoint_raw(self) -> int:  # noqa R0201
        """Return raw heat set point(number only, no units)."""
        return util.BOGUS_INT  # placeholder

    def get_heat_setpoint(self) -> int:  # noqa R0201
        """Return raw heat set point(number and units)."""
        return util.BOGUS_INT  # placeholder

    def get_schedule_program_heat(self) -> dict:  # noqa R0201
        """
        Return the heat setpoint schedule.

        inputs:
            None
        returns:
            (dict): scheduled heat set points and times in degrees.
        """
        return util.bogus_dict  # placeholder

    def get_schedule_heat_sp(self) -> int:
        """Return the heat setpoint."""
        return util.BOGUS_INT  # placeholder

    def get_cool_setpoint_raw(self) -> int:  # noqa R0201
        """Return raw cool set point (number only, no units)."""
        return util.BOGUS_INT  # placeholder

    def get_cool_setpoint(self) -> int:  # noqa R0201
        """Return raw cool set point (number and units)."""
        return util.BOGUS_INT  # placeholder

    def get_schedule_program_cool(self) -> dict:  # noqa R0201
        """
        Return the cool setpoint schedule.

        inputs:
            None
        returns:
            (dict): scheduled cool set points and times in degrees.
        """
        return util.bogus_dict  # placeholder

    def get_schedule_cool_sp(self) -> int:
        """Return the cool setpoint."""
        return util.BOGUS_INT  # placeholder

    def get_vacation_hold(self) -> bool:  # noqa R0201
        """Return True if thermostat is in vacation hold mode."""
        return util.BOGUS_BOOL  # placeholder

    def get_is_invacation_hold_mode(self) -> bool:  # noqa R0201
        """Return the 'IsInVacationHoldMode' setting."""
        return util.BOGUS_BOOL  # placeholder

    def get_temporary_hold_until_time(self) -> int:  # noqa R0201
        """Return the 'TemporaryHoldUntilTime' """
        return util.BOGUS_INT  # placeholder

    def refresh_zone_info(self, force_refresh=False) -> None:
        """
        Refresh zone info.

        inputs:
            force_refresh(bool): if True, ignore expiration timer.
        returns:
            None, cached data is refreshed.
        """
        del force_refresh  # not used in this template.
        self.zone_info = {}
        self.last_fetch_time = time.time()

    def report_heating_parameters(self, switch_position=None):  # noqa R0201
        """
        Display critical thermostat settings and reading to the screen.

        inputs:
            switch_position(int): switch position override for testing.
        returns:
            None
        """
        del switch_position  # placeholder

    def get_vacation_hold_until_time(self) -> int:  # noqa R0201
        """
        Return the 'VacationHoldUntilTime'.

        inputs:
            None
        returns:
            (int): vacation hold time in minutes.
         """
        return util.BOGUS_INT  # not implemented

    def update_runtime_parameters(self):
        """use runtime parameter overrides.

        inputs:
            None
        returns:
            None, updates class variables.
        """
        # map user input keys to class members.
        # "thermostat_type is not overwritten
        user_input_to_class_mapping = {
            api.input_flds.thermostat_type: "thermostat_type",
            api.input_flds.zone: "zone_name",
            api.input_flds.poll_time: "poll_time_sec",
            api.input_flds.connection_time: "connection_time_sec",
            api.input_flds.tolerance: "tolerance_degrees",
            api.input_flds.target_mode: "target_mode",
            api.input_flds.measurements: "measurements",
        }

        if self.verbose:
            print("\n")
            util.log_msg("supervisor runtime parameters:",
                         mode=util.BOTH_LOG, func_name=1)
        for inp, cls_method in user_input_to_class_mapping.items():
            user_input = api.uip.get_user_inputs(api.uip.zone_name, inp)
            if user_input is not None:
                setattr(self, cls_method, user_input)
                if self.verbose:
                    util.log_msg(f"{inp}={user_input}",
                                 mode=util.BOTH_LOG, func_name=1)

    def verify_current_mode(self, target_mode):
        """
        Verify current mode matches target mode.

        inputs:
            target_mode(str): target mode override
        returns:
            (bool): True if current mode matches target mode,
                    or target mode is not specified.
        """
        if target_mode is None:
            return True
        else:
            return bool(self.current_mode == target_mode)

    def revert_thermostat_mode(self, target_mode):
        """
        Revert thermostat mode to target mode.

        inputs:
            target_mode(str): target mode override
        returns:
            target_mode(str) target_mode, which may get updated by
            this function.
        """
        # do not switch directly from hot to cold
        if (self.current_mode in self.heat_modes and
                target_mode in self.cool_modes):
            util.log_msg(
                f"WARNING: target mode={target_mode}, switching from "
                f"{self.current_mode} to OFF_MODE to prevent damage to HVAC",
                mode=util.BOTH_LOG,
                func_name=1)
            target_mode = self.OFF_MODE

        # do not switch directly from cold to hot
        elif (self.current_mode in self.cool_modes and
              target_mode in self.heat_modes):
            util.log_msg(
                f"WARNING: target mode={target_mode}, switching from "
                f"{self.current_mode} to OFF_MODE to prevent damage to HVAC",
                mode=util.BOTH_LOG,
                func_name=1)
            target_mode = self.OFF_MODE

        # revert the mode to target
        # UNKNOWN mode is bypassed
        if target_mode == self.UNKNOWN_MODE:
            print(f"{util.get_function_name()}: target_mode='{target_mode}', "
                  "doing nothing.")
        else:
            self.set_mode(target_mode)

        return target_mode

    def measure_thermostat_response_time(self, measurements=30,
                                         func=None):
        """
        Measure Thermostat response time and report statistics.

        inputs:
            measurements(int): number of measurements
            func(obj): target function to run during timing measurement.
        returns:
            (dict): measurement statistics.
        """
        delta_lst = []
        stats = {}
        # set default measurement method if not provided.
        if func is None:
            func = self.get_display_temp

        # measurement loop
        for measurement in range(measurements):
            time0 = time.time()
            func()  # target command
            time1 = time.time()

            # accumulate stats
            tdelta = time1 - time0
            delta_lst.append(tdelta)
            util.log_msg(f"measurement {measurement}={tdelta:.2f} seconds",
                         mode=util.BOTH_LOG, func_name=1)

        # calc stats
        stats["measurements"] = measurements
        stats["mean"] = round(statistics.mean(delta_lst), 2)
        stats["stdev"] = round(statistics.stdev(delta_lst), 2)
        stats["min"] = round(min(delta_lst), 2)
        stats["max"] = round(max(delta_lst), 2)
        stats["3sigma_upper"] = round((3.0 * stats["stdev"] +
                                      stats["mean"]), 2)
        stats["6sigma_upper"] = round((6.0 * stats["stdev"] +
                                       stats["mean"]), 2)
        return stats

    def measure_thermostat_repeatability(self, measurements=30,
                                         poll_interval_sec=0,
                                         func=None):
        """
        Measure Thermostat repeatability and report statistics.

        inputs:
            measurements(int): number of measurements
            poll_interval_sec(int): delay between measurements
            func(obj): target function to run during repeatabilty measurement.
        returns:
            (dict): measurement statistics.
        """
        data_lst = []
        stats = {}
        # set default measurement method if not provided.
        if func is None:
            func = self.get_display_temp

        # measurement loop
        for measurement in range(measurements):
            t_start = time.time()
            data = func()  # target command
            t_end = time.time()

            # accumulate stats
            tdelta = t_end - t_start
            data_lst.append(data)
            util.log_msg(
                f"measurement {measurement}={data} (deltaTime={tdelta:.2f} "
                f"sec, delay={poll_interval_sec} sec)",
                mode=util.BOTH_LOG,
                func_name=1)
            time.sleep(poll_interval_sec)
            self.refresh_zone_info()

        # calc stats
        stats["measurements"] = measurements
        stats["mean"] = round(statistics.mean(data_lst), 2)
        stats["stdev"] = round(statistics.stdev(data_lst), 2)
        stats["min"] = round(min(data_lst), 2)
        stats["max"] = round(max(data_lst), 2)
        stats["3sigma_upper"] = round((3.0 * stats["stdev"] +
                                      stats["mean"]), 2)
        stats["6sigma_upper"] = round((6.0 * stats["stdev"] +
                                       stats["mean"]), 2)
        return stats

    def display_basic_thermostat_summary(self, mode=util.CONSOLE_LOG):
        """
        Display basic thermostat summary.

        inputs:
            mode(int): target log for data.
        returns:
            None, prints data to log and/or console.
        """
        print("\n")
        util.log_msg("current thermostat settings...",
                     mode=mode, func_name=1)
        util.log_msg(f"zone name='{self.zone_name}'",
                     mode=mode, func_name=1)
        sw_pos = util.get_key_from_value(self.system_switch_position,
                                         self.get_system_switch_position())
        util.log_msg(
            f"system switch position: {self.get_system_switch_position()} "
            f"({sw_pos})",
            mode=mode,
            func_name=1)
        util.log_msg(
            f"thermostat display temp="
            f"{util.temp_value_with_units(self.get_display_temp())}",
            mode=mode,
            func_name=1)
        util.log_msg(
            f"thermostat display humidity="
            f"{util.humidity_value_with_units(self.get_display_humidity())}",
            mode=mode,
            func_name=1)
        util.log_msg(
            f"heat set point="
            f"{util.temp_value_with_units(self.get_heat_setpoint())}",
            mode=mode,
            func_name=1)
        util.log_msg(
            f"cool set point="
            f"{util.temp_value_with_units(self.get_cool_setpoint())}",
            mode=mode,
            func_name=1)
        util.log_msg(
            f"heat schedule set point="
            f"{util.temp_value_with_units(self.get_schedule_heat_sp())}",
            mode=mode,
            func_name=1)
        util.log_msg(
            f"cool schedule set point="
            f"{util.temp_value_with_units(self.get_schedule_cool_sp())}",
            mode=mode,
            func_name=1)
        util.log_msg(
            f"(schedule) heat program={self.get_schedule_program_heat()}",
            mode=mode, func_name=1)
        util.log_msg(
            f"(schedule) cool program={self.get_schedule_program_cool()}",
            mode=mode, func_name=1)
        util.log_msg(
            f"heat mode={self.is_heat_mode()} "
            f"(actively heating={self.is_heating()})",
            mode=mode,
            func_name=1)
        util.log_msg(
            f"cool mode={self.is_cool_mode()} "
            f"(actively cooling={self.is_cooling()})",
            mode=mode,
            func_name=1)
        util.log_msg(
            f"dry mode={self.is_dry_mode()} "
            f"(actively drying={self.is_drying()})",
            mode=mode,
            func_name=1)
        util.log_msg(
            f"auto mode={self.is_auto_mode()} "
            f"(actively auto={self.is_auto()})",
            mode=mode,
            func_name=1)
        util.log_msg(
            f"fan mode={self.is_fan_mode()} "
            f"(actively fanning={self.is_fanning()})",
            mode=mode,
            func_name=1)
        util.log_msg(f"off mode={self.is_off_mode()}",
                     mode=mode, func_name=1)
        util.log_msg(f"hold={self.get_vacation_hold()}",
                     mode=mode, func_name=1)
        util.log_msg(
            f"temporary hold minutes={self.get_temporary_hold_until_time()}",
            mode=mode, func_name=1)

    def revert_temperature_deviation(self, setpoint=None, msg=""):
        """
        Revert the temperature deviation.

        inputs:
            setpoint(int or float): setpoint in deg F
            msg(str): status message to display.
        returns:
            None
        """
        if setpoint is None:
            setpoint = self.current_setpoint

        eml.send_email_alert(
            subject=f"{self.thermostat_type} {self.current_mode.upper()} "
            f"deviation alert on zone {self.zone_name}",
            body=msg)
        util.log_msg(
            f"\n*** {self.thermostat_type} {self.current_mode.upper()} "
            f"deviation detected on zone {self.zone_name}, "
            f"reverting thermostat to heat schedule ***\n",
            mode=util.BOTH_LOG)
        self.revert_setpoint_func(setpoint)

    def function_not_supported(self, *_, **__):
        """Function for unsupported activities."""
        util.log_msg(
            f"WARNING (in {util.get_function_name(2)}): function call is "
            f"not supported on this thermostat type",
            mode=util.BOTH_LOG)

    def display_runtime_settings(self):
        """
        Display runtime settings to console.

        inputs:
            None
        returns:
            None
        """
        # poll time setting:
        util.log_msg(
            f"polling time set to {self.poll_time_sec / 60.0:.1f} minutes",
            mode=util.BOTH_LOG)

        # reconnection time to thermostat server:
        util.log_msg(
            f"server re-connect time set to "
            f"{self.connection_time_sec / 60.0:.1f}"
            f" minutes",
            mode=util.BOTH_LOG)

        # tolerance to set point:
        util.log_msg(
            f"tolerance to set point is set to "
            f"{util.temp_value_with_units(self.tolerance_degrees)}",
            mode=util.BOTH_LOG)

    def display_session_settings(self):
        """
        Display session settings to console.

        inputs:
            None
        returns:
            None
        """
        # set log file name
        util.log_msg.file_name = (self.thermostat_type + "_" +
                                  str(self.zone_name) + ".txt")

        util.log_msg(
            f"{self.thermostat_type} thermostat zone {self.zone_name} "
            f"monitoring service\n",
            mode=util.BOTH_LOG)

        util.log_msg("session settings:", mode=util.BOTH_LOG)

        util.log_msg("thermostat %s for %s\n" %
                     (["is being monitored", "will be reverted"]
                      [self.revert_deviations],
                      ["energy consuming deviations\n("
                       "e.g. heat setpoint above schedule "
                       "setpoint, cool setpoint below schedule"
                       " setpoint)",
                       "all schedule deviations"]
                      [self.revert_all_deviations]), mode=util.BOTH_LOG)

    def supervisor_loop(self, Thermostat, session_count, measurement,
                        debug):
        """
        Loop through supervisor algorithm.

        inputs:
            Thermostat(obj):  Thermostat instance object
            session_count(int):  current session
            measurement(int):  current measurement index
            debug(bool): debug flag
        returns:
            measurement(int): current measurement count
        """
        global connection_ok  # noqa W603

        # initialize poll counter
        poll_count = 1
        previous_mode_dict = {}

        # poll thermostat settings
        while not api.uip.max_measurement_count_exceeded(measurement):
            # query thermostat for current settings and set points
            current_mode_dict = self.get_current_mode(
                session_count, poll_count,
                flag_all_deviations=self.revert_all_deviations)

            # debug data on change from previous poll
            # note this check is probably hyper-sensitive, since status msg
            # change could trigger this extra report.
            if current_mode_dict != previous_mode_dict:
                if debug:
                    self.report_heating_parameters()
                previous_mode_dict = current_mode_dict  # latch

            # revert thermostat mode if not matching target
            if not self.verify_current_mode(api.uip.get_user_inputs(
                    api.uip.zone_name, api.input_flds.target_mode)):
                api.uip.set_user_inputs(api.uip.zone_name,
                                        api.input_flds.target_mode,
                                        self.revert_thermostat_mode(
                                            api.uip.get_user_inputs(
                                                api.uip.zone_name,
                                                api.input_flds.target_mode)
                                            ))

            # revert thermostat to schedule if heat override is detected
            if (self.revert_deviations and self.is_controlled_mode() and
                    self.is_temp_deviated_from_schedule()):
                self.revert_temperature_deviation(
                    self.schedule_setpoint, current_mode_dict["status_msg"])

            # increment poll count
            poll_count += 1
            measurement += 1

            # polling delay
            time.sleep(self.poll_time_sec)

            # refresh zone info
            connection_ok = True
            self.refresh_zone_info()

            # reconnect
            if ((time.time() - self.session_start_time_sec)
                    > self.connection_time_sec) or not connection_ok:
                util.log_msg("forcing re-connection to thermostat...",
                             mode=util.BOTH_LOG)
                del Thermostat
                break  # force reconnection

        # return measurement count
        return measurement


def create_thermostat_instance(thermostat_type, zone,
                               ThermostatClass, ThermostatZone,
                               verbose=True):
    """
    Create Thermostat and Zone instances.

    inputs:
        tstat(int):  thermostat_type
        zone(int): zone number
        zone_name(str): name of zone
        ThermostatClass(cls): Thermostat class
        ThermostatZone(cls): ThermostatZone class
        verbose(bool): debug flag
    returns:
        Thermostat(obj): Thermostat object
        Zone(obj):  Zone object
    """
    util.log_msg.debug = verbose  # debug mode set

    # verify required env vars
    api.verify_required_env_variables(thermostat_type, str(zone),
                                      verbose=verbose)

    # import hardware module
    api.load_hardware_library(thermostat_type)

    # create Thermostat object
    Thermostat = ThermostatClass(zone, verbose=verbose)
    if verbose:
        Thermostat.print_all_thermostat_metadata(zone)

    # create Zone object
    Zone = ThermostatZone(Thermostat, verbose=verbose)

    # update runtime overrides
    # thermostat_type
    api.uip.set_user_inputs(api.uip.zone_name,
                            api.input_flds.thermostat_type,
                            thermostat_type)
    # zone
    api.uip.set_user_inputs(api.uip.zone_name, api.input_flds.zone,
                            zone)
    Zone.update_runtime_parameters()

    return Thermostat, Zone


def thermostat_basic_checkout(thermostat_type, zone,
                              ThermostatClass, ThermostatZone):
    """
    Perform basic Thermostat checkout.

    inputs:
        tstat(int):  thermostat_type
        zone(int): zone number
        ThermostatClass(cls): Thermostat class
        ThermostatZone(cls): ThermostatZone class
    returns:
        Thermostat(obj): Thermostat object
        Zone(obj):  Zone object
    """
    util.log_msg.debug = True  # debug mode set

    # create class instances
    Thermostat, Zone = create_thermostat_instance(thermostat_type, zone,
                                                  ThermostatClass,
                                                  ThermostatZone)

    Zone.display_basic_thermostat_summary()

    return Thermostat, Zone


def thermostat_get_all_zone_temps(thermostat_type, zone_lst,
                                  ThermostatClass, ThermostatZone):
    """
    Perform basic Thermostat checkout.

    inputs:
        tstat(int):  thermostat_type
        zone_lst(list): list of zones
        ThermostatClass(cls): Thermostat class
        ThermostatZone(cls): ThermostatZone class
    returns:
        Thermostat(obj): Thermostat object
        Zone(obj):  Zone object
    """
    util.log_msg.debug = False  # debug mode unset
    print("\nquerying all zone temps:")

    for zone in zone_lst:
        # create class instances
        Thermostat, Zone = create_thermostat_instance(thermostat_type, zone,
                                                      ThermostatClass,
                                                      ThermostatZone,
                                                      verbose=False)

        display_temp = Zone.get_display_temp()
        zone_name = Zone.zone_name
        print(f"zone: {zone}, name: {zone_name}, temp: {display_temp}")

    return Thermostat, Zone



"""
Integration test module for emulator.py.

This test requires connection to emulator thermostat.
"""
# built-in imports
import math
import unittest

# local imports
from thermostatsupervisor import emulator
from thermostatsupervisor import emulator_config
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


class IntegrationTest(utc.IntegrationTest):
    """
    Test functions in emulator.py.
    """

    def setUpIntTest(self):
        """Setup common to integration tests."""
        self.setup_common()
        self.print_test_name()

        # emulator argv list must be valid settings
        self.unit_test_argv = [
            "supervise.py",  # module
            "emulator",  # thermostat type
            "0",  # zone
            "5",  # poll time in sec, this value violates min
            # cycle time for TCC if reverting temperature deviation
            "11",  # reconnect time in sec
            "2",  # tolerance
            "UNKNOWN_MODE",  # thermostat mode, no target
            "6",  # number of measurements
        ]
        self.mod = emulator
        self.mod_config = emulator_config


class FunctionalIntegrationTest(IntegrationTest,
                                utc.FunctionalIntegrationTest):
    """
    Test functional performance of emulator.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()
        # test_GetMetaData input parameters
        self.metadata_field = "display_temp"
        self.metadata_type = float


class SuperviseIntegrationTest(IntegrationTest,
                               utc.SuperviseIntegrationTest):
    """
    Test supervise functionality of emulator.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()


class PerformanceIntegrationTest(IntegrationTest,
                                 utc.PerformanceIntegrationTest):
    """
    Test performance of in emulator.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()
        # network timing measurement
        self.timeout_limit = 1
        self.timing_measurements = 30  # fast measurement

        # temperature and humidity repeatability measurements
        # data is uniform distribution, 1 std= (range)/sqrt(12)
        # adding 0.1 buffer to prevent false failures due to rounding, etc.
        self.temp_stdev_limit = \
            emulator_config.NORMAL_TEMP_VARIATION * 2 / math.sqrt(12) + 0.1
        self.temp_repeatability_measurements = 100  # number of temp msmts.
        self.humidity_stdev_limit = \
            emulator_config.NORMAL_HUMIDITY_VARIATION * 2 / math.sqrt(12) + 0.1
        self.humidity_repeatability_measurements = 100  # number of temp msmts.
        self.poll_interval_sec = 0.5  # delay between repeatability msmts


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""
Integration test module for honeywell.py.

This test requires connection to Honeywell thermostat.
"""
# built-in imports
import unittest

# local imports
from thermostatsupervisor import honeywell
from thermostatsupervisor import honeywell_config
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


class IntegrationTest(utc.IntegrationTest):
    """
    Test functions in honeywell.py.
    """

    def setUpIntTest(self):
        """Setup common to integration tests."""
        self.setup_common()
        self.print_test_name()

        # Honeywell argv list must be valid settings
        self.unit_test_argv = [
            "supervise.py",  # module
            "honeywell",  # thermostat
            "0",  # zone
            "30",  # poll time in sec, this value violates min
            # cycle time for TCC if reverting temperature deviation
            "1000",  # reconnect time in sec
            "2",  # tolerance
            "UNKNOWN_MODE",  # thermostat mode, no target
            "3",  # number of measurements
        ]
        self.mod = honeywell
        self.mod_config = honeywell_config


class FunctionalIntegrationTest(IntegrationTest,
                                utc.FunctionalIntegrationTest):
    """
    Test functional performance of honeywell.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()
        # test_GetMetaData input parameters
        self.metadata_field = "DeviceID"
        self.metadata_type = int


class SuperviseIntegrationTest(IntegrationTest,
                               utc.SuperviseIntegrationTest):
    """
    Test supervise functionality of honeywell.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()


class PerformanceIntegrationTest(IntegrationTest,
                                 utc.PerformanceIntegrationTest):
    """
    Test performance of in honeywell.py.
    """

    def setUp(self):
        super().setUp()
        self.setUpIntTest()
        # network timing measurement
        self.timeout_limit = honeywell.HTTP_TIMEOUT
        self.timing_measurements = 30  # fast measurement

        # temperature and humidity repeatability measurements
        # TCC server polling period to thermostat appears to be about 5-6 min
        # temperature and humidity data are int values
        # settings below are tuned for 12 minutes, 4 measurements per minute.
        self.temp_stdev_limit = 0.5  # 1 sigma temp repeatability limit in F
        self.temp_repeatability_measurements = 48  # number of temp msmts.
        self.humidity_stdev_limit = 0.5  # 1 sigma humid repeat. limit %RH
        self.humidity_repeatability_measurements = 48  # number of temp msmts.
        self.poll_interval_sec = 15  # delay between repeatability measurements


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""
Unit test module for sht31_flask_server.py.

Flask server tests currently do not work on Azure pipelines
because ports cannot be opened on shared pool.
"""
# built-in imports
import os
import unittest

# local imports
# thermostat_api is imported but not used to avoid a circular import
from thermostatsupervisor import environment as env
from thermostatsupervisor import thermostat_api as api  # noqa F401, pylint: disable=unused-import.
from thermostatsupervisor import sht31
from thermostatsupervisor import sht31_config
from thermostatsupervisor import sht31_flask_server as sht31_fs
from thermostatsupervisor import utilities as util
from tests import unit_test_common as utc


@unittest.skipIf(not utc.ENABLE_SHT31_TESTS,
                 "sht31 tests are disabled")
@unittest.skipIf(env.is_azure_environment(),
                 "this test not supported on Azure Pipelines")
@unittest.skipIf(not utc.ENABLE_FLASK_INTEGRATION_TESTS,
                 "flask integration tests are disabled")
class IntegrationTest(utc.UnitTest):
    """Test functions in sht31_flask_server.py."""

    # sht31 flask server is automatically spawned in sht31
    # Thermostat class if unit test zone is being used.

    def test_sht31_flask_server_all_pages(self):
        """
        Confirm all pages return data from Flask server.
        """
        # do not test these pages
        no_test_list = ["i2c_recovery", "reset"]

        # loopback does not work so use local sht31 zone if testing
        # on the local net.  If not, use the DNS name.
        zone = sht31_config.get_preferred_zone()

        for test_case in sht31_config.flask_folder:
            if test_case in no_test_list:
                print(f"test_case={test_case}: bypassing this test case")
                continue

            print(f"test_case={test_case}")
            Thermostat = \
                sht31.ThermostatClass(
                    zone, path=sht31_config.flask_folder[test_case])
            print("printing thermostat meta data:")
            return_data = Thermostat.print_all_thermostat_metadata(
                zone)

            # validate dictionary was returned
            self.assertTrue(isinstance(return_data, dict),
                            "return data is not a dictionary")

            # validate key as proof of correct return page
            if test_case in ["production", "unit_test"]:
                expected_key = "measurements"
            elif test_case in ["diag", "clear_diag", "enable_heater",
                               "disable_heater", "soft_reset"]:
                expected_key = "raw_binary"
            elif test_case in ["i2c_detect", "i2c_detect_0", "i2c_detect_1"]:
                expected_key = "i2c_detect"
            elif test_case == "i2c_recovery":
                expected_key = "i2c_recovery"
            elif test_case == "reset":
                expected_key = "message"
            else:
                expected_key = "bogus"
            self.assertTrue(expected_key in return_data,
                            f"test_case '{test_case}': key '{expected_key}' "
                            f"was not found in return data: {return_data}")

    def test_sht31_flask_server(self):
        """
        Confirm Flask server returns valid data.
        """
        measurements_bckup = sht31_config.MEASUREMENTS
        try:
            for sht31_config.measurements in [1, 10, 100, 1000]:
                msg = ['measurement', 'measurements'][
                    sht31_config.MEASUREMENTS > 1]
                print(f"\ntesting SHT31 flask server with "
                      f"{sht31_config.MEASUREMENTS} {msg}...")
                self.validate_flask_server()
        finally:
            sht31_config.measurements = measurements_bckup

    def validate_flask_server(self):
        """
        Launch SHT31 Flask server and validate data.
        """
        print("creating thermostat object...")
        Thermostat = sht31.ThermostatClass(sht31_config.UNIT_TEST_ZONE)
        print("printing thermostat meta data:")
        Thermostat.print_all_thermostat_metadata(sht31_config.UNIT_TEST_ZONE)

        # create mock runtime args
        api.uip = api.UserInputs(utc.unit_test_sht31)

        # create Zone object
        Zone = sht31.ThermostatZone(Thermostat)

        # update runtime overrides
        Zone.update_runtime_parameters()

        print("current thermostat settings...")
        print(f"switch position: {Zone.get_system_switch_position()}")
        print(f"heat mode={Zone.is_heat_mode()}")
        print(f"cool mode={Zone.is_cool_mode()}")
        print(f"temporary hold minutes={Zone.get_temporary_hold_until_time()}")
        meta_data = Thermostat.get_all_metadata(sht31_config.UNIT_TEST_ZONE)
        print(f"thermostat meta data={meta_data}")
        print(f"thermostat display temp="
              f"{util.temp_value_with_units(Zone.get_display_temp())}")

        # verify measurements
        self.assertEqual(meta_data["measurements"],
                         sht31_config.MEASUREMENTS,
                         f"measurements: actual={meta_data['measurements']}, "
                         f"expected={sht31_config.MEASUREMENTS}")

        # verify metadata
        test_cases = {
            "get_display_temp": {"min_val": 80, "max_val": 120},
            "get_is_humidity_supported": {"min_val": True, "max_val": True},
            "get_display_humidity": {"min_val": 49, "max_val": 51},
        }
        for param, limits in test_cases.items():
            return_val = getattr(Zone, param)()
            print(f"'{param}'={return_val}")
            min_val = limits["min_val"]
            max_val = limits["max_val"]
            self.assertTrue(min_val <= return_val <= max_val,
                            f"'{param}'={return_val}, not between {min_val} "
                            f"and {max_val}")
        # cleanup
        del Zone
        del Thermostat


class RuntimeParameterTest(utc.RuntimeParameterTest):
    """sht31 flask server Runtime parameter tests."""

    mod = sht31_fs  # module to test
    script = os.path.realpath(__file__)
    debug = False

    # fields for testing, mapped to class variables.
    # (value, field name)
    test_fields = [
        (script, os.path.realpath(__file__)),
        (debug, sht31_fs.input_flds.debug_fld),
    ]


if __name__ == "__main__":
    util.log_msg.debug = True
    unittest.main(verbosity=2)



"""
Thermostat API.

This file should be updated for any new thermostats supported and
any changes to thermostat configs.
"""
# built ins
import munch

# local imports
from thermostatsupervisor import blink_config
from thermostatsupervisor import emulator_config
from thermostatsupervisor import honeywell_config
from thermostatsupervisor import kumocloud_config
from thermostatsupervisor import kumolocal_config
from thermostatsupervisor import mmm_config
from thermostatsupervisor import sht31_config
from thermostatsupervisor import environment as env
from thermostatsupervisor import utilities as util

# thermostat types
DEFAULT_THERMOSTAT = emulator_config.ALIAS
DEFAULT_ZONE_NAME = util.default_parent_key

# list of thermostat config modules supported
config_modules = [blink_config,
                  emulator_config,
                  honeywell_config,
                  kumocloud_config,
                  kumolocal_config,
                  mmm_config,
                  sht31_config
                  ]

SUPPORTED_THERMOSTATS = {
    # "module" = module to import
    # "type" = thermostat type index number
    # "zones" = zone numbers supported
    # "modes" = modes supported
}
for config_module in config_modules:
    SUPPORTED_THERMOSTATS.update(
        {config_module.ALIAS: config_module.supported_configs})

# dictionary of required env variables for each thermostat type
thermostats = {
}
for config_module in config_modules:
    thermostats.update(
        {config_module.ALIAS: {"required_env_variables":
                               config_module.required_env_variables}})


# runtime override parameters
# note script name is omitted, starting with first parameter
# index 0 (script name) is not included in this dict because it is
# not a runtime argument
input_flds = munch.Munch()
input_flds.thermostat_type = "thermostat_type"
input_flds.zone = "zone"
input_flds.poll_time = "poll_time"
input_flds.connection_time = "connection_time"
input_flds.tolerance = "tolerance"
input_flds.target_mode = "target_mode"
input_flds.measurements = "measurements"
input_flds.input_file = "input_file"

uip = None  # user inputs object


class UserInputs(util.UserInputs):
    """Manage runtime arguments for thermostat_api."""

    def __init__(self, argv_list=None, help_description=None,
                 suppress_warnings=False, thermostat_type=DEFAULT_THERMOSTAT,
                 zone_name=DEFAULT_ZONE_NAME):
        """
        UserInputs constructor for thermostat_api.

        inputs:
            argv_list(list): override runtime values.
            help_description(str): description field for help text.
            suppress_warnings(bool): True to suppress warning msgs.
            thermostat_type(str): thermostat type.
            zone_name(str): thermostat zone name (e.g. 'living room')
        """
        self.argv_list = argv_list
        self.help_description = help_description
        self.suppress_warnings = suppress_warnings
        self.thermostat_type = thermostat_type  # default if not provided
        self.zone_name = zone_name
        print(f"DEBUG ({util.get_function_name()}: "
              f"self.zone_name={self.zone_name}")

        # initialize parent class
        super().__init__(argv_list, help_description, suppress_warnings,
                         zone_name)

    def initialize_user_inputs(self, parent_keys=None):
        """
        Populate user_inputs dict.

        inputs:
            parent_keys(list): list of parent keys
        """
        if parent_keys is None:
            parent_keys = [self.default_parent_key]
        self.valid_sflags = []
        self.user_inputs = {}  # init
        # define the user_inputs dict.
        for parent_key in parent_keys:
            self.user_inputs[parent_key] = {
                input_flds.thermostat_type: {
                    "order": 1,  # index in the argv list
                    "value": None,
                    "type": str,
                    "default": self.thermostat_type,
                    "valid_range": list(SUPPORTED_THERMOSTATS.keys()),
                    "sflag": "-t",
                    "lflag": "--" + input_flds.thermostat_type,
                    "help": "thermostat type",
                    "required": False,  # default value is set if missing.
                    },
                input_flds.zone: {
                    "order": 2,  # index in the argv list
                    "value": None,
                    "type": int,
                    "default": 0,
                    "valid_range": None,  # updated once thermostat is known
                    "sflag": "-z",
                    "lflag": "--" + input_flds.zone,
                    "help": "target zone number",
                    "required": False,  # defaults to idx 0 in supported zones
                    },
                input_flds.poll_time: {
                    "order": 3,  # index in the argv list
                    "value": None,
                    "type": int,
                    "default": 60 * 10,
                    "valid_range": range(0, 24 * 60 * 60),
                    "sflag": "-p",
                    "lflag": "--" + input_flds.poll_time,
                    "help": "poll time (sec)",
                    "required": False,
                    },
                input_flds.connection_time: {
                    "order": 4,  # index in the argv list
                    "value": None,
                    "type": int,
                    "default": 24 * 60 * 60,  # 1 day
                    "valid_range": range(0, 7 * 24 * 60 * 60),  # up to 1 wk
                    "sflag": "-c",
                    "lflag": "--" + input_flds.connection_time,
                    "help": "server connection time (sec)",
                    "required": False,
                    },
                input_flds.tolerance: {
                    "order": 5,  # index in the argv list
                    "value": None,
                    "type": int,
                    "default": 2,
                    "valid_range": range(0, 10),
                    "sflag": "-d",
                    "lflag": "--" + input_flds.tolerance,
                    "help": "tolerance (deg F)",
                    "required": False,
                    },
                input_flds.target_mode: {
                    "order": 6,  # index in the argv list
                    "value": None,
                    "type": str,
                    "default": "UNKNOWN_MODE",
                    "valid_range": None,  # updated once thermostat is known
                    "sflag": "-m",
                    "lflag": "--" + input_flds.target_mode,
                    "help": "target thermostat mode",
                    "required": False,
                    },
                input_flds.measurements: {
                    "order": 7,  # index in the argv list
                    "value": None,
                    "type": int,
                    "default": 10000,
                    "valid_range": range(1, 10001),
                    "sflag": "-n",
                    "lflag": "--" + input_flds.measurements,
                    "help": "number of measurements",
                    "required": False,
                    },
                input_flds.input_file: {
                    "order": 8,  # index in the argv list
                    "value": None,
                    # "type": lambda x: self.is_valid_file(x),
                    "type": str,  # argparse.FileType('r', encoding='UTF-8'),
                    "default": None,
                    "valid_range": None,
                    "sflag": "-f",
                    "lflag": "--" + input_flds.input_file,
                    "help": "input file",
                    "required": False,
                    },
            }
            self.valid_sflags += [self.user_inputs[parent_key][k]["sflag"]
                                  for k in self.user_inputs[parent_key].keys()]

    def dynamic_update_user_inputs(self):
        """
        Update thermostat-specific values in user_inputs dict.

        This function expands each input parameter list to match
        the length of the thermostat parameter field.
        """
        # initializ section list to single item list of one thermostat
        self.parent_keys = [self.default_parent_key]  # initialize

        # file input will override any type of individual inputs
        input_file = self.get_user_inputs(self.default_parent_key,
                                          input_flds.input_file)
        if input_file is not None:
            # aise Exception("dynamic_update_user_inputs found")
            self.using_input_file = True
            print(f"DEBUG: reading runtime arguments from '{input_file}'...")
            self.parse_input_file(input_file)
            # scan all sections in INI file in reversed order so that
            # user_inputs contains the first key after casting.
            # section = list(self.user_inputs_file.keys())[0]  # use first key
            self.parent_keys = list(self.user_inputs_file.keys())
            # reinit user_inputs dict based on INI file structure
            self.initialize_user_inputs(self.parent_keys)
            # populate user_inputs from user_inputs_file
            for section in self.parent_keys:
                for fld in input_flds:
                    if fld == input_flds.input_file:
                        # input file field will not be in the file
                        continue
                    if (self.user_inputs[section][fld]["type"]
                            in [int, float, str]):
                        # cast data type when reading value
                        try:
                            self.user_inputs[section][fld]["value"] = (
                                self.user_inputs[section][fld]["type"](
                                    self.user_inputs_file[section].get(
                                        input_flds[fld])))
                        except Exception:
                            print(f"exception in section={section}, fld={fld}")
                            raise
                        # cast original input value in user_inputs_file as well
                        self.user_inputs_file[section][input_flds[fld]] = \
                            self.user_inputs[section][fld]["value"]
                    else:
                        # no casting, just read raw from list
                        self.user_inputs[section][fld]["value"] = \
                            ((self.user_inputs_file[section].get(input_flds[
                                fld])))
            # for now set zone name to first zone in file
            self.zone_name = self.parent_keys[0]
            print(f"DEBUG ({util.get_function_name()}: "
                  f"self.zone_name={self.zone_name}")
        # update user_inputs parent_key with zone_name
        # if user_inputs has already been populated
        elif (self.get_user_inputs(list(self.user_inputs.keys())[0],
                                   input_flds.thermostat_type) is not None):
            # argv inputs, only currenty supporting 1 zone
            # verify only 1 parent key exists
            current_keys = list(self.user_inputs.keys())
            if len(current_keys) != 1:
                raise KeyError(f"user_input keys={current_keys}, expected only"
                               " 1 key")

            # update parent key to be zone_name
            current_key = current_keys[0]
            new_key = (self.get_user_inputs(
                       current_key, input_flds.thermostat_type) + "_" +
                       str(self.get_user_inputs(current_key, input_flds.zone)))
            self.user_inputs[new_key] = self.user_inputs.pop(current_key)

            # update paremeters for new parent keys
            self.zone_name = new_key  # set Zone name
            print(f"DEBUG ({util.get_function_name()}: "
                  f"self.zone_name={self.zone_name}")
            self.default_parent_key = new_key
            self.parent_keys = list(self.user_inputs.keys())
        else:
            runtime_args = (self.get_user_inputs(list(
                self.user_inputs.keys())[0], input_flds.thermostat_type))
            print(f"runtime args: {runtime_args}")

        # if thermostat is not set yet, default it based on module
        # TODO - code block needs update for multi-zone
        for zone_name in self.parent_keys:
            thermostat_type = self.get_user_inputs(zone_name,
                                                   input_flds.thermostat_type)
            if thermostat_type is None:
                thermostat_type = self.thermostat_type
            try:
                self.user_inputs[zone_name][input_flds.zone]["valid_range"] = \
                    SUPPORTED_THERMOSTATS[thermostat_type]["zones"]
            except KeyError:
                print(f"\nKeyError: one or more keys are invalid (zone_name="
                      f"{zone_name}, zone_number={input_flds.zone}, "
                      f"thermostat_type={thermostat_type})\n")
                raise
            try:
                self.user_inputs[zone_name][input_flds.target_mode][
                    "valid_range"] = \
                    SUPPORTED_THERMOSTATS[thermostat_type]["modes"]
            except KeyError:
                print(f"\nKeyError: one or more keys are invalid (zone_name="
                      f"{zone_name}, target_mode={input_flds.target_mode}, "
                      f"thermostat_type={thermostat_type})\n")
                raise

    def max_measurement_count_exceeded(self, measurement):
        """
        Return True if max measurement reached.

        inputs:
            measurement(int): current measurement value
        returns:
            (bool): True if max measurement reached.
        """
        max_measurements = self.get_user_inputs(self.zone_name, "measurements")
        if max_measurements is None:
            return False
        elif measurement > max_measurements:
            return True
        else:
            return False


def verify_required_env_variables(tstat, zone_str, verbose=True):
    """
    Verify all required env variables are present for thermostat
    configuration in use.

    inputs:
        tstat(int) thermostat type mapping to thermostat_api
        zone_str(str): zone input as a string
        verbose(bool): debug flag.
    returns:
        (bool): True if all keys are present, else False
    """
    if verbose:
        print("\nchecking required environment variables:")
    key_status = True  # default, all keys present
    for key in thermostats[tstat]["required_env_variables"]:
        # any env key ending in '_' should have zone number appended to it.
        if key[-1] == '_':
            # append zone info to key
            key = key + str(zone_str)
        if verbose:
            print(f"checking required environment key: {key}...", end='')
        env.env_variables[key] = env.get_env_variable(key)["value"]
        if env.env_variables[key] is not None:
            if verbose:
                print("OK")
        else:
            util.log_msg(
                f"{tstat}: zone {zone_str}: FATAL error: one or more required"
                f" environemental keys are missing, exiting program",
                mode=util.BOTH_LOG)
            key_status = False
            raise KeyError
    if verbose:
        print("\n")
    return key_status


def load_hardware_library(thermostat_type):
    """
    Dynamic load 3rd party library for requested hardware type.

    inputs:
        thermostat_type(str): thermostat alias string
    returns:
        (obj): loaded python module
    """
    pkg_name = (util.PACKAGE_NAME + "." +
                SUPPORTED_THERMOSTATS[thermostat_type]["module"])
    mod = env.dynamic_module_import(pkg_name)
    return mod


def load_user_inputs(config_mod):
    """
    Load the default user inputs and return the zone number.

    inputs:
        config_mod(obj): config module
    returns:
        zone_number(int): zone number
    """
    global uip
    zone_name = config_mod.default_zone_name
    print(f"DEBUG in load user inputs, zone_name={zone_name}")
    uip = UserInputs(argv_list=config_mod.argv,
                     thermostat_type=config_mod.ALIAS,
                     zone_name=zone_name)
    zone_number = uip.get_user_inputs(uip.zone_name,
                                      input_flds.zone)
    return zone_number



"""Generic Flask functionality."""

# built-in libraries
import datetime

# third party libraries
from flask_apscheduler import APScheduler
from flask_ipban import IpBan

# local imports

# ipban
ipban_ban_count = 1
ipban_ban_seconds = 3600 * 24 * 7  # 1 wk
ipban_persistent = False  # True to persist across restarts


def initialize_ipban(app):
    """
    Initialize ipban agent for blocking hacking attempts.

    inputs:
        app (flask app object)
    returns:
        (ip_ban object)
    """
    # setup ipban
    ip_ban = IpBan(app=app,
                   ban_count=ipban_ban_count,
                   ban_seconds=ipban_ban_seconds,
                   persist=ipban_persistent)
    ip_ban.init_app(app)
    ip_ban.load_nuisances()
    print_ipban_block_list(ip_ban)
    return ip_ban


def print_ipban_block_list(ip_ban):
    """
    Print the current ip_ban block list to the console.

    inputs:
        ip_ban(ip_ban object)
    returns:
        None
    """
    # s = ""
    # s += "<table class='table'><thead>\n"
    # s += ("<tr><th>ip</th><th>count</th><th>permanent</th><th>url</th><th>"
    #       f"timestamp</th></tr>\n")
    # s += ""</thead><tbody>\n"
    # for k, r in ip_ban.get_block_list().items():
    #     s += (f"<tr><td>{k}</td><td>{r['count']}</td><td>"
    #           f"{r.get('permanent', '')}</td><td>{r.get('url', '')}</td><td>"
    #           f"{r['timestamp']}</td></tr>\n")
    # print(f"{s}")
    print(f"ip_ban black list: {ip_ban.get_block_list()}")


def print_ipban_block_list_with_timestamp(ip_ban):
    """
    Print the current ip_ban block list to the console with timestamp.

    inputs:
        ip_ban(ip_ban object)
    returns:
        None
    """
    now = datetime.datetime.now()
    now_str = now.strftime("%Y-%m-%d %H:%M:%S")
    print(f"{now_str}: ip_ban black list: {ip_ban.get_block_list()}")


def schedule_ipban_block_list_report(ip_ban, debug_mode=False):
    """
    Schedule an ip_ban blocked ip list report.

    inputs:
        ip_ban(ip_ban object)
    returns:
        None
    """
    # interval = 1 day std, 1 min in debug mode
    interval_sec = 60 * [60 * 24, 1][debug_mode]
    print(f"ip_ban blacklist report scheduled every {interval_sec / 60.0} "
          "minutes")
    scheduler = APScheduler()
    kwargs = {"ip_ban": ip_ban}
    scheduler.add_job(id='ip_ban blacklist report',
                      func=print_ipban_block_list_with_timestamp,
                      kwargs=kwargs,
                      trigger="interval", seconds=interval_sec)
    scheduler.start()


def set_flask_cookie_config(app):
    """
    Set cookie config to protect against cookie attack vectors in
    the Flask configuration.

    inputs:
        app (flask app object)
    returns:
        None
    """
    app.config.update(
        SESSION_COOKIE_SECURE=True,
        SESSION_COOKIE_HTTPONLY=True,
        SESSION_COOKIE_SAMESITE='Lax',
    )


def print_flask_config(app):
    """
    """
    print("flask config:")
    print(f"{app.config}")
